@@ -30,26 +30,35 @@
 
 #include "fs-rtp-codec-cache.h"
 
+#include "fs-rtp-conference.h"
+
 #include <gst/farsight/fs-conference-iface.h>
 
 #include <string.h>
-#include <unistd.h>
-
-#ifdef STANDALONE
-# undef GST_WARNING
-# undef GST_DEBUG
-# define GST_DEBUG(...) g_debug (__VA_ARGS__)
-# define GST_WARNING(...) g_warning (__VA_ARGS__)
-#else
-# include "fs-rtp-conference.h"
-# define GST_CAT_DEFAULT fsrtpconference_disco
+#ifdef HAVE_UNISTD_H
+# include <unistd.h>
 #endif
 
-static gboolean codecs_cache_valid(gchar *cache_path) {
+/* Because of annoying CRTs */
+#if defined (_MSC_VER) && _MSC_VER >= 1400
+# include <io.h>
+# define open _open
+# define close _close
+# define read _read
+# define write _write
+# define stat _stat
+# define STAT_TYPE struct _stat
+#else
+# define STAT_TYPE struct stat
+#endif
+
+#define GST_CAT_DEFAULT fsrtpconference_disco
+
+static gboolean codecs_cache_valid (gchar *cache_path) {
   time_t cache_ts = 0;
   time_t registry_ts = 0;
-  struct stat cache_stat;
-  struct stat registry_stat;
+  STAT_TYPE cache_stat;
+  STAT_TYPE registry_stat;
   gchar *registry_xml_path;
   gchar *registry_bin_path;
 
@@ -60,31 +69,31 @@ static gboolean codecs_cache_valid(gchar *cache_path) {
     registry_xml_path = g_build_filename (g_get_home_dir (),
         ".gstreamer-" GST_MAJORMINOR, "registry." HOST_CPU ".xml", NULL);
   } else {
-    registry_bin_path = g_strdup(registry_xml_path);
+    registry_bin_path = g_strdup (registry_xml_path);
   }
 
-  if (stat(registry_xml_path, &registry_stat) == 0) {
+  if (stat (registry_xml_path, &registry_stat) == 0) {
     registry_ts = registry_stat.st_mtime;
   }
 
-  if (stat(registry_bin_path, &registry_stat) == 0) {
+  if (stat (registry_bin_path, &registry_stat) == 0) {
     if (registry_ts < registry_stat.st_mtime) {
       registry_ts = registry_stat.st_mtime;
     }
   }
 
-  if (stat(cache_path, &cache_stat) == 0) {
+  if (stat (cache_path, &cache_stat) == 0) {
     cache_ts = cache_stat.st_mtime;
   }
 
-  g_free(registry_bin_path);
-  g_free(registry_xml_path);
+  g_free (registry_bin_path);
+  g_free (registry_xml_path);
 
   return (registry_ts != 0 && cache_ts > registry_ts);
 }
 
 static gchar *
-get_codecs_cache_path (FsMediaType media_type, GError **error) {
+get_codecs_cache_path (FsMediaType media_type) {
   gchar *cache_path;
 
   if (media_type == FS_MEDIA_TYPE_AUDIO) {
@@ -93,15 +102,14 @@ get_codecs_cache_path (FsMediaType media_type, GError **error) {
       cache_path = g_build_filename (g_get_home_dir (), ".farsight",
           "codecs.audio." HOST_CPU ".cache", NULL);
     }
-  } else if(media_type == FS_MEDIA_TYPE_VIDEO) {
+  } else if (media_type == FS_MEDIA_TYPE_VIDEO) {
     cache_path = g_strdup (g_getenv ("FS_VIDEO_CODECS_CACHE"));
     if (cache_path == NULL) {
       cache_path = g_build_filename (g_get_home_dir (), ".farsight",
           "codecs.video." HOST_CPU ".cache", NULL);
     }
   } else {
-    g_set_error (error, FS_ERROR, FS_ERROR_INVALID_ARGUMENTS,
-      "Unknown media type %d for cache loading", media_type);
+    GST_ERROR ("Unknown media type %d for cache loading", media_type);
     return NULL;
   }
 
@@ -111,23 +119,23 @@ get_codecs_cache_path (FsMediaType media_type, GError **error) {
 
 static gboolean
 read_codec_blueprint_uint (gchar **in, gsize *size, guint *val) {
-  if (*size < sizeof(guint))
+  if (*size < sizeof (guint))
     return FALSE;
 
   *val = *((guint *) *in);
-  *in += sizeof(guint);
-  *size -= sizeof(guint);
+  *in += sizeof (guint);
+  *size -= sizeof (guint);
   return TRUE;
 }
 
 static gboolean
 read_codec_blueprint_int (gchar **in, gsize *size, gint *val) {
-  if (*size < sizeof(gint))
+  if (*size < sizeof (gint))
     return FALSE;
 
   *val = *((gint *) *in);
-  *in += sizeof(gint);
-  *size -= sizeof(gint);
+  *in += sizeof (gint);
+  *size -= sizeof (gint);
   return TRUE;
 }
 
@@ -150,16 +158,16 @@ read_codec_blueprint_string (gchar **in, gsize *size, gchar **str) {
 }
 
 
-#define READ_CHECK(x) if(!x) goto error;
+#define READ_CHECK(x) if (!x) goto error;
 
 static CodecBlueprint *
 load_codec_blueprint (FsMediaType media_type, gchar **in, gsize *size) {
-  CodecBlueprint *codec_blueprint = g_new0 (CodecBlueprint, 1);
+  CodecBlueprint *codec_blueprint = g_slice_new0 (CodecBlueprint);
   gchar *tmp;
   gint tmp_size;
   int i;
 
-  codec_blueprint->codec = g_new0 (FsCodec, 1);
+  codec_blueprint->codec = g_slice_new0 (FsCodec);
   codec_blueprint->codec->media_type = media_type;
 
   READ_CHECK (read_codec_blueprint_int
@@ -173,12 +181,13 @@ load_codec_blueprint (FsMediaType media_type, gchar **in, gsize *size) {
 
 
   READ_CHECK (read_codec_blueprint_int (in, size, &tmp_size));
-  for(i = 0; i < tmp_size; i++) {
-    FsCodecParameter *param = g_new0 (FsCodecParameter, 1);
-    READ_CHECK (read_codec_blueprint_string (in, size, &(param->name)));
-    READ_CHECK (read_codec_blueprint_string (in, size, &(param->value)));
-    codec_blueprint->codec->optional_params =
-        g_list_append (codec_blueprint->codec->optional_params, param);
+  for (i = 0; i < tmp_size; i++) {
+    gchar *name, *value;
+    READ_CHECK (read_codec_blueprint_string (in, size, &(name)));
+    READ_CHECK (read_codec_blueprint_string (in, size, &(value)));
+    fs_codec_add_optional_parameter (codec_blueprint->codec, name, value);
+    g_free (name);
+    g_free (value);
   }
 
   READ_CHECK (read_codec_blueprint_string (in, size, &tmp));
@@ -190,7 +199,7 @@ load_codec_blueprint (FsMediaType media_type, gchar **in, gsize *size) {
   g_free (tmp);
 
   READ_CHECK (read_codec_blueprint_int (in, size, &tmp_size));
-  for(i = 0; i < tmp_size; i++) {
+  for (i = 0; i < tmp_size; i++) {
     int j, tmp_size2;
     GList *tmplist = NULL;
 
@@ -207,7 +216,7 @@ load_codec_blueprint (FsMediaType media_type, gchar **in, gsize *size) {
   }
 
   READ_CHECK (read_codec_blueprint_int (in, size, &tmp_size));
-  for(i = 0; i < tmp_size; i++) {
+  for (i = 0; i < tmp_size; i++) {
     int j, tmp_size2;
     GList *tmplist = NULL;
 
@@ -247,7 +256,7 @@ load_codec_blueprint (FsMediaType media_type, gchar **in, gsize *size) {
  *
  */
 GList *
-load_codecs_cache (FsMediaType media_type, GError **error)
+load_codecs_cache (FsMediaType media_type)
 {
   GMappedFile *mapped = NULL;
   gchar *contents = NULL;
@@ -265,23 +274,22 @@ load_codecs_cache (FsMediaType media_type, GError **error)
 
   if (media_type == FS_MEDIA_TYPE_AUDIO) {
     magic_media = 'A';
-  } else if(media_type == FS_MEDIA_TYPE_VIDEO) {
+  } else if (media_type == FS_MEDIA_TYPE_VIDEO) {
     magic_media = 'V';
   } else {
-    g_set_error (error, FS_ERROR, FS_ERROR_INVALID_ARGUMENTS,
-      "Invalid media type %d", media_type);
-    return FALSE;
+    GST_ERROR ("Invalid media type %d", media_type);
+    return NULL;
   }
 
-  cache_path = get_codecs_cache_path(media_type, error);
+  cache_path = get_codecs_cache_path (media_type);
 
   if (!cache_path)
-    return FALSE;
+    return NULL;
 
-  if (!codecs_cache_valid(cache_path)) {
+  if (!codecs_cache_valid (cache_path)) {
     GST_DEBUG ("Codecs cache %s is outdated or does not exist", cache_path);
-    g_free(cache_path);
-    return FALSE;
+    g_free (cache_path);
+    return NULL;
   }
 
   GST_DEBUG ("Loading codecs cache %s", cache_path);
@@ -292,12 +300,11 @@ load_codecs_cache (FsMediaType media_type, GError **error)
       err ? err->message: "unknown error");
     g_clear_error (&err);
 
-    if(!g_file_get_contents (cache_path, &contents, &size, error))
+    if (!g_file_get_contents (cache_path, &contents, &size, NULL))
       goto error;
   } else {
     if ((contents = g_mapped_file_get_contents (mapped)) == NULL) {
-      g_set_error (error, FS_ERROR, FS_ERROR_INTERNAL,
-        "Can't load file %s : %s", cache_path, g_strerror (errno));
+      GST_WARNING ("Can't load file %s : %s", cache_path, g_strerror (errno));
       goto error;
     }
     /* check length for header */
@@ -309,13 +316,13 @@ load_codecs_cache (FsMediaType media_type, GError **error)
   in = contents;
 
   if (size < sizeof (magic)) {
-    g_set_error (error, FS_ERROR, FS_ERROR_INTERNAL, "Cache file corrupt");
+    GST_WARNING ("Cache file corrupt");
     goto error;
   }
 
-  memcpy (magic, in, sizeof(magic));
-  in += sizeof(magic);
-  size -= sizeof(magic);
+  memcpy (magic, in, sizeof (magic));
+  in += sizeof (magic);
+  size -= sizeof (magic);
 
   if (magic[0] != 'F' ||
       magic[1] != 'S' ||
@@ -323,26 +330,24 @@ load_codecs_cache (FsMediaType media_type, GError **error)
       magic[3] != 'C' ||
       magic[4] != '1' ||   /* This is the version number */
       magic[5] != '1') {
-    g_set_error (error, FS_ERROR, FS_ERROR_INTERNAL,
-      "Cache file has incorrect magic header. File corrupted");
+    GST_WARNING ("Cache file has incorrect magic header. File corrupted");
     goto error;
   }
 
-  if (size < sizeof(gint)) {
-    g_set_error (error, FS_ERROR, FS_ERROR_INTERNAL,
-      "Cache file corrupt (size: %"G_GSIZE_FORMAT" < sizeof(int))", size);
+  if (size < sizeof (gint)) {
+    GST_WARNING ("Cache file corrupt (size: %"G_GSIZE_FORMAT" < sizeof (int))",
+        size);
     goto error;
   }
 
   num_blueprints = *((gint *) in);
-  in += sizeof(gint);
-  size -= sizeof(gint);
+  in += sizeof (gint);
+  size -= sizeof (gint);
 
   for (i = 0; i < num_blueprints; i++) {
     CodecBlueprint *blueprint = load_codec_blueprint (media_type, &in, &size);
     if (!blueprint) {
-      g_set_error (error, FS_ERROR, FS_ERROR_INTERNAL,
-        "Can not load all of the blueprints, cache corrupted");
+      GST_WARNING ("Can not load all of the blueprints, cache corrupted");
 
       if (blueprints) {
         g_list_foreach (blueprints, (GFunc) codec_blueprint_destroy, NULL);
@@ -352,7 +357,7 @@ load_codecs_cache (FsMediaType media_type, GError **error)
 
       goto error;
     }
-    blueprints = g_list_prepend (blueprints, blueprint);
+    blueprints = g_list_append (blueprints, blueprint);
   }
 
  error:
@@ -365,11 +370,11 @@ load_codecs_cache (FsMediaType media_type, GError **error)
   return blueprints;
 }
 
-#define WRITE_CHECK(x) if(!x) return FALSE;
+#define WRITE_CHECK(x) if (!x) return FALSE;
 
 static gboolean
 write_codec_blueprint_int (int fd, gint val) {
-  return write (fd, &val, sizeof(gint)) == sizeof(gint);
+  return write (fd, &val, sizeof (gint)) == sizeof (gint);
 }
 
 static gboolean
@@ -416,13 +421,13 @@ save_codec_blueprint (int fd, CodecBlueprint *codec_blueprint) {
 
   walk = codec_blueprint->send_pipeline_factory;
   size = g_list_length (walk);
-  if (write (fd, &size, sizeof(gint)) != sizeof(gint))
+  if (write (fd, &size, sizeof (gint)) != sizeof (gint))
     return FALSE;
 
   for (; walk; walk = g_list_next (walk)) {
     GList *walk2 = walk->data;
     size = g_list_length (walk2);
-    if (write (fd, &size, sizeof(gint)) != sizeof(gint))
+    if (write (fd, &size, sizeof (gint)) != sizeof (gint))
       return FALSE;
     for (; walk2; walk2 = g_list_next (walk2)) {
       GstElementFactory *fact = walk2->data;
@@ -433,13 +438,13 @@ save_codec_blueprint (int fd, CodecBlueprint *codec_blueprint) {
 
   walk = codec_blueprint->receive_pipeline_factory;
   size = g_list_length (walk);
-  if (write (fd, &size, sizeof(gint)) != sizeof(gint))
+  if (write (fd, &size, sizeof (gint)) != sizeof (gint))
     return FALSE;
 
   for (; walk; walk = g_list_next (walk)) {
     GList *walk2 = walk->data;
     size = g_list_length (walk2);
-    if (write (fd, &size, sizeof(gint)) != sizeof(gint))
+    if (write (fd, &size, sizeof (gint)) != sizeof (gint))
       return FALSE;
     for (; walk2; walk2 = g_list_next (walk2)) {
       GstElementFactory *fact = walk2->data;
@@ -453,7 +458,7 @@ save_codec_blueprint (int fd, CodecBlueprint *codec_blueprint) {
 
 
 gboolean
-save_codecs_cache(FsMediaType media_type, GList *blueprints)
+save_codecs_cache (FsMediaType media_type, GList *blueprints)
 {
   gchar *cache_path;
   GList *item;
@@ -462,7 +467,7 @@ save_codecs_cache(FsMediaType media_type, GList *blueprints)
   int size;
   gchar magic[8] = {0};
 
-  cache_path = get_codecs_cache_path(media_type, NULL);
+  cache_path = get_codecs_cache_path (media_type);
   if (!cache_path)
     return FALSE;
 
@@ -485,7 +490,7 @@ save_codecs_cache(FsMediaType media_type, GList *blueprints)
     fd = g_mkstemp (tmp_path);
 
     if (fd == -1) {
-      GST_DEBUG ("Unable to save codecs cache. g_mkstemp() failed: %s",
+      GST_DEBUG ("Unable to save codecs cache. g_mkstemp () failed: %s",
           g_strerror (errno));
       g_free (tmp_path);
       g_free (cache_path);
@@ -500,7 +505,7 @@ save_codecs_cache(FsMediaType media_type, GList *blueprints)
 
   if (media_type == FS_MEDIA_TYPE_AUDIO) {
     magic[2] = 'A';
-  } else if(media_type == FS_MEDIA_TYPE_VIDEO) {
+  } else if (media_type == FS_MEDIA_TYPE_VIDEO) {
     magic[2] = 'V';
   }
 
@@ -513,7 +518,7 @@ save_codecs_cache(FsMediaType media_type, GList *blueprints)
 
 
   size = g_list_length (blueprints);
-  if (write (fd, &size, sizeof(gint)) != sizeof(gint))
+  if (write (fd, &size, sizeof (gint)) != sizeof (gint))
     return FALSE;
 
 
@@ -546,6 +551,7 @@ save_codecs_cache(FsMediaType media_type, GList *blueprints)
   }
 
   g_free (tmp_path);
+  g_free (cache_path);
   GST_DEBUG ("Wrote binary codecs cache");
   return TRUE;
 }
