@@ -26,7 +26,9 @@
  * SECTION:fs-rtp-stream
  * @short_description: A RTP stream in a #FsRtpSession in a #FsRtpConference
  *
-
+ * This is the conjunction of a #FsRtpParticipant and a #FsRtpSession,
+ * it is created by calling fs_session_new_stream() on a
+ * #FsRtpSession.
  */
 
 #ifdef HAVE_CONFIG_H
@@ -52,6 +54,7 @@ enum
   PROP_SOURCE_PADS,
 #endif
   PROP_REMOTE_CODECS,
+  PROP_NEGOTIATED_CODECS,
   PROP_CURRENT_RECV_CODECS,
   PROP_DIRECTION,
   PROP_PARTICIPANT,
@@ -62,20 +65,17 @@ enum
 struct _FsRtpStreamPrivate
 {
   FsRtpSession *session;
-  FsRtpParticipant *participant;
   FsStreamTransmitter *stream_transmitter;
 
   FsStreamDirection direction;
 
-  /* Protected by the session mutex */
-  GList *substreams;
-  guint recv_codecs_changed_idle_id;
-
   GError *construction_error;
 
-  GList *known_ssrcs;
+  stream_new_remote_codecs_cb new_remote_codecs_cb;
+  stream_known_source_packet_receive_cb known_source_packet_received_cb;
+  gpointer user_data_for_cb;
 
-  gboolean disposed;
+  GMutex *mutex;
 };
 
 
@@ -98,14 +98,12 @@ static void fs_rtp_stream_set_property (GObject *object,
 static void fs_rtp_stream_constructed (GObject *object);
 
 
-static gboolean fs_rtp_stream_add_remote_candidate (FsStream *stream,
-                                                    FsCandidate *candidate,
-                                                    GError **error);
-static void fs_rtp_stream_remote_candidates_added (FsStream *stream);
-static gboolean fs_rtp_stream_select_candidate_pair (FsStream *stream,
-                                                     gchar *lfoundation,
-                                                     gchar *rfoundatihon,
+static gboolean fs_rtp_stream_set_remote_candidates (FsStream *stream,
+                                                     GList *candidates,
                                                      GError **error);
+static gboolean fs_rtp_stream_force_remote_candidates (FsStream *stream,
+    GList *remote_candidates,
+    GError **error);
 
 static gboolean fs_rtp_stream_set_remote_codecs (FsStream *stream,
                                                  GList *remote_codecs,
@@ -123,13 +121,23 @@ static void _new_local_candidate (
     FsStreamTransmitter *stream_transmitter,
     FsCandidate *candidate,
     gpointer user_data);
+static void
+_known_source_packet_received (FsStreamTransmitter *st,
+    guint component,
+    GstBuffer *buffer,
+    FsRtpStream *self);
 static void _transmitter_error (
     FsStreamTransmitter *stream_transmitter,
     gint errorno,
     gchar *error_msg,
     gchar *debug_msg,
     gpointer user_data);
-
+static void _substream_codec_changed (FsRtpSubStream *substream,
+    FsRtpStream *stream);
+static void _state_changed (FsStreamTransmitter *stream_transmitter,
+    guint component,
+    FsStreamState state,
+    gpointer user_data);
 
 
 static GObjectClass *parent_class = NULL;
@@ -147,21 +155,22 @@ fs_rtp_stream_class_init (FsRtpStreamClass *klass)
   gobject_class->set_property = fs_rtp_stream_set_property;
   gobject_class->get_property = fs_rtp_stream_get_property;
   gobject_class->constructed = fs_rtp_stream_constructed;
+  gobject_class->dispose = fs_rtp_stream_dispose;
+  gobject_class->finalize = fs_rtp_stream_finalize;
 
-  stream_class->add_remote_candidate = fs_rtp_stream_add_remote_candidate;
+  stream_class->set_remote_candidates = fs_rtp_stream_set_remote_candidates;
   stream_class->set_remote_codecs = fs_rtp_stream_set_remote_codecs;
-  stream_class->remote_candidates_added = fs_rtp_stream_remote_candidates_added;
-  stream_class->select_candidate_pair = fs_rtp_stream_select_candidate_pair;
+  stream_class->force_remote_candidates = fs_rtp_stream_force_remote_candidates;
 
-#if 0
-  g_object_class_override_property (gobject_class,
-                                    PROP_SOURCE_PADS,
-                                    "source-pads");
-#endif
+
+  g_type_class_add_private (klass, sizeof (FsRtpStreamPrivate));
 
   g_object_class_override_property (gobject_class,
                                     PROP_REMOTE_CODECS,
                                     "remote-codecs");
+  g_object_class_override_property (gobject_class,
+                                    PROP_NEGOTIATED_CODECS,
+                                    "negotiated-codecs");
   g_object_class_override_property (gobject_class,
                                     PROP_CURRENT_RECV_CODECS,
                                     "current-recv-codecs");
@@ -177,12 +186,6 @@ fs_rtp_stream_class_init (FsRtpStreamClass *klass)
   g_object_class_override_property (gobject_class,
                                     PROP_STREAM_TRANSMITTER,
                                    "stream-transmitter");
-
-
-  gobject_class->dispose = fs_rtp_stream_dispose;
-  gobject_class->finalize = fs_rtp_stream_finalize;
-
-  g_type_class_add_private (klass, sizeof (FsRtpStreamPrivate));
 }
 
 static void
@@ -191,58 +194,102 @@ fs_rtp_stream_init (FsRtpStream *self)
   /* member init */
   self->priv = FS_RTP_STREAM_GET_PRIVATE (self);
 
-  self->priv->disposed = FALSE;
   self->priv->session = NULL;
-  self->priv->participant = NULL;
+  self->participant = NULL;
   self->priv->stream_transmitter = NULL;
-  self->priv->known_ssrcs = NULL;
+
+  self->priv->mutex = g_mutex_new ();
 
   self->priv->direction = FS_DIRECTION_NONE;
 }
 
+static FsRtpSession *
+fs_rtp_stream_get_session (FsRtpStream *self, GError **error)
+{
+  FsRtpSession *session;
+
+  g_mutex_lock (self->priv->mutex);
+  session = self->priv->session;
+  if (session)
+    g_object_ref (session);
+  g_mutex_unlock (self->priv->mutex);
+
+  if (!session)
+    g_set_error (error, FS_ERROR, FS_ERROR_DISPOSED,
+        "Called function after stream has been disposed");
+
+  return session;
+}
+
+
+static FsStreamTransmitter *
+fs_rtp_stream_get_stream_transmitter (FsRtpStream *self, GError **error)
+{
+  FsRtpSession *session = fs_rtp_stream_get_session (self, error);
+  FsStreamTransmitter *st = NULL;
+
+  if (!session)
+    return NULL;
+
+  FS_RTP_SESSION_LOCK (session);
+  st = self->priv->stream_transmitter;
+  if (st)
+    g_object_ref (st);
+  FS_RTP_SESSION_UNLOCK (session);
+
+  if (!st)
+    g_set_error (error, FS_ERROR, FS_ERROR_DISPOSED,
+        "Called function after stream has been disposed");
+
+  g_object_unref (session);
+  return st;
+}
+
 static void
 fs_rtp_stream_dispose (GObject *object)
 {
   FsRtpStream *self = FS_RTP_STREAM (object);
+  FsStreamTransmitter *st;
+  FsRtpParticipant *participant;
+  FsRtpSession *session = fs_rtp_stream_get_session (self, NULL);
 
-  if (self->priv->disposed) {
-    /* If dispose did already run, return. */
+  if (!session)
     return;
-  }
 
-  if (self->priv->stream_transmitter) {
-    g_object_unref (self->priv->stream_transmitter);
-    self->priv->stream_transmitter = NULL;
-  }
+  g_mutex_lock (self->priv->mutex);
+  self->priv->session = NULL;
+  g_mutex_unlock (self->priv->mutex);
 
-  FS_RTP_SESSION_LOCK (self->priv->session);
-  if (self->priv->recv_codecs_changed_idle_id)
+  FS_RTP_SESSION_LOCK (session);
+  participant = self->participant;
+  self->participant = NULL;
+
+  st = self->priv->stream_transmitter;
+  self->priv->stream_transmitter = NULL;
+
+  if (st)
   {
-    g_source_remove (self->priv->recv_codecs_changed_idle_id);
-    self->priv->recv_codecs_changed_idle_id = 0;
+    FS_RTP_SESSION_UNLOCK (session);
+    fs_stream_transmitter_stop (st);
+    g_object_unref (st);
+    FS_RTP_SESSION_LOCK (session);
   }
 
-  if (self->priv->substreams) {
-    g_list_foreach (self->priv->substreams, (GFunc) g_object_unref, NULL);
-    g_list_free (self->priv->substreams);
-    self->priv->substreams = NULL;
-  }
-  FS_RTP_SESSION_UNLOCK (self->priv->session);
-
-  if (self->priv->participant) {
-    g_object_unref (self->priv->participant);
-    self->priv->participant = NULL;
-  }
-
-  /* Make sure dispose does not run twice. */
-  self->priv->disposed = TRUE;
-
-  if (self->priv->session)
+  while (self->substreams)
   {
-    g_object_unref (self->priv->session);
-    self->priv->session = NULL;
+    FsRtpSubStream *substream = self->substreams->data;
+    self->substreams = g_list_remove (self->substreams, substream);
+    FS_RTP_SESSION_UNLOCK (session);
+    g_object_unref (substream);
+    FS_RTP_SESSION_LOCK (session);
   }
 
+  FS_RTP_SESSION_UNLOCK (session);
+
+  g_object_unref (participant);
+  g_object_unref (session);
+  g_object_unref (session);
+
   parent_class->dispose (object);
 }
 
@@ -251,11 +298,10 @@ fs_rtp_stream_finalize (GObject *object)
 {
   FsRtpStream *self = FS_RTP_STREAM (object);
 
-  if (self->remote_codecs)
-    fs_codec_list_destroy (self->remote_codecs);
+  fs_codec_list_destroy (self->remote_codecs);
+  fs_codec_list_destroy (self->negotiated_codecs);
 
-  if (self->priv->known_ssrcs)
-    g_list_free (self->priv->known_ssrcs);
+  g_mutex_free (self->priv->mutex);
 
   parent_class->finalize (object);
 }
@@ -280,19 +326,29 @@ fs_rtp_stream_get_property (GObject *object,
                             GParamSpec *pspec)
 {
   FsRtpStream *self = FS_RTP_STREAM (object);
+  FsRtpSession *session = fs_rtp_stream_get_session (self, NULL);
+
+  if (!session)
+    return;
 
   switch (prop_id) {
     case PROP_REMOTE_CODECS:
+      FS_RTP_SESSION_LOCK (session);
       g_value_set_boxed (value, self->remote_codecs);
+      FS_RTP_SESSION_UNLOCK (session);
+      break;
+    case PROP_NEGOTIATED_CODECS:
+      FS_RTP_SESSION_LOCK (session);
+      g_value_set_boxed (value, self->negotiated_codecs);
+      FS_RTP_SESSION_UNLOCK (session);
       break;
     case PROP_SESSION:
-      g_value_set_object (value, self->priv->session);
+      g_value_set_object (value, session);
       break;
     case PROP_PARTICIPANT:
-      g_value_set_object (value, self->priv->participant);
-      break;
-    case PROP_STREAM_TRANSMITTER:
-      g_value_set_object (value, self->priv->stream_transmitter);
+      FS_RTP_SESSION_LOCK (session);
+      g_value_set_object (value, self->participant);
+      FS_RTP_SESSION_UNLOCK (session);
       break;
     case PROP_DIRECTION:
       g_value_set_flags (value, self->priv->direction);
@@ -302,25 +358,31 @@ fs_rtp_stream_get_property (GObject *object,
         GList *codeclist = NULL;
         GList *substream_item;
 
-        FS_RTP_SESSION_LOCK (self->priv->session);
-        for (substream_item = g_list_first (self->priv->substreams);
+        FS_RTP_SESSION_LOCK (session);
+        for (substream_item = g_list_first (self->substreams);
              substream_item;
              substream_item = g_list_next (substream_item))
         {
-          FsCodec *codec = NULL;
-          g_object_get (substream_item->data, "codec", &codec, NULL);
-          if (!_codec_list_has_codec (codeclist, codec))
-            codeclist = g_list_append (codeclist, codec);
+          FsRtpSubStream *substream = substream_item->data;
+
+          if (substream->codec)
+          {
+            if (!_codec_list_has_codec (codeclist, substream->codec))
+              codeclist = g_list_append (codeclist,
+                  fs_codec_copy (substream->codec));
+          }
         }
 
         g_value_take_boxed (value, codeclist);
-        FS_RTP_SESSION_UNLOCK (self->priv->session);
+        FS_RTP_SESSION_UNLOCK (session);
       }
       break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
   }
+
+  g_object_unref (session);
 }
 
 static void
@@ -337,30 +399,55 @@ fs_rtp_stream_set_property (GObject *object,
       self->priv->session = FS_RTP_SESSION (g_value_dup_object (value));
       break;
     case PROP_PARTICIPANT:
-      self->priv->participant = FS_RTP_PARTICIPANT (g_value_dup_object (value));
+      self->participant = FS_RTP_PARTICIPANT (g_value_dup_object (value));
       break;
     case PROP_STREAM_TRANSMITTER:
       self->priv->stream_transmitter =
         FS_STREAM_TRANSMITTER (g_value_get_object (value));
       break;
     case PROP_DIRECTION:
-      self->priv->direction = g_value_get_flags (value);
-      if (self->priv->stream_transmitter)
-        g_object_set (self->priv->stream_transmitter, "sending",
-            self->priv->direction & FS_DIRECTION_SEND, NULL);
-      FS_RTP_SESSION_LOCK (self->priv->session);
-      for (item = g_list_first (self->priv->substreams);
-           item;
-           item = g_list_next (item))
-        g_object_set (G_OBJECT (item->data),
-            "receiving", ((self->priv->direction & FS_DIRECTION_RECV) != 0),
-            NULL);
-      FS_RTP_SESSION_UNLOCK (self->priv->session);
+      {
+        FsStreamTransmitter *st = NULL;
+        GList *copy = NULL;
+        FsRtpSession *session = fs_rtp_stream_get_session (self, NULL);
+        FsStreamDirection dir;
+
+        if (!session)
+        {
+          self->priv->direction = g_value_get_flags (value);
+          return;
+        }
+
+        FS_RTP_SESSION_LOCK (session);
+        dir = self->priv->direction = g_value_get_flags (value);
+        FS_RTP_SESSION_UNLOCK (session);
+        st = fs_rtp_stream_get_stream_transmitter (self, NULL);
+        if (st)
+        {
+          g_object_set (self->priv->stream_transmitter, "sending",
+              dir & FS_DIRECTION_SEND, NULL);
+          g_object_unref (st);
+        }
+
+        FS_RTP_SESSION_LOCK (session);
+        copy = g_list_copy (g_list_first (self->substreams));
+        g_list_foreach (copy, (GFunc) g_object_ref, NULL);
+        FS_RTP_SESSION_UNLOCK (session);
+
+        for (item = copy;  item; item = g_list_next (item))
+          g_object_set (G_OBJECT (item->data),
+              "receiving", ((dir & FS_DIRECTION_RECV) != 0),
+              NULL);
+        g_list_foreach (copy, (GFunc) g_object_unref, NULL);
+        g_list_free (copy);
+        g_object_unref (session);
+      }
       break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
   }
+
 }
 
 static void
@@ -394,69 +481,76 @@ fs_rtp_stream_constructed (GObject *object)
       "error",
       G_CALLBACK (_transmitter_error),
       self);
+  g_signal_connect (self->priv->stream_transmitter,
+      "known-source-packet-received",
+      G_CALLBACK (_known_source_packet_received),
+      self);
+  g_signal_connect (self->priv->stream_transmitter,
+      "state-changed",
+      G_CALLBACK (_state_changed),
+      self);
+
+  if (!fs_stream_transmitter_gather_local_candidates (
+          self->priv->stream_transmitter,
+          &self->priv->construction_error))
+  {
+    if (!self->priv->construction_error)
+      self->priv->construction_error = g_error_new (FS_ERROR,
+          FS_ERROR_INTERNAL,
+          "Unknown error while gathering local candidates");
+    return;
+  }
+
+  GST_CALL_PARENT (G_OBJECT_CLASS, constructed, (object));
 }
 
 
 /**
- * fs_rtp_stream_add_remote_candidate:
- * @stream: an #FsStream
- * @candidate: an #FsCandidate struct representing a remote candidate
- * @error: location of a #GError, or NULL if no error occured
- *
- * This function adds the given candidate into the remote candiate list of the
- * stream. It will be used for establishing a connection with the peer. A copy
- * will be made so the user must free the passed candidate using
- * fs_candidate_destroy() when done.
+ * fs_rtp_stream_set_remote_candidate:
  */
 static gboolean
-fs_rtp_stream_add_remote_candidate (FsStream *stream, FsCandidate *candidate,
-                                    GError **error)
+fs_rtp_stream_set_remote_candidates (FsStream *stream, GList *candidates,
+                                     GError **error)
 {
   FsRtpStream *self = FS_RTP_STREAM (stream);
+  FsStreamTransmitter *st = fs_rtp_stream_get_stream_transmitter (self, error);
+  gboolean ret = FALSE;
 
-  return fs_stream_transmitter_add_remote_candidate (
-      self->priv->stream_transmitter, candidate, error);
-}
+  if (!st)
+    return FALSE;
 
+  ret = fs_stream_transmitter_set_remote_candidates (st, candidates, error);
 
-/**
- * fs_rtp_stream_remote_candidates_added:
- * @stream: a #FsStream
- *
- * Call this function when the remotes candidates have been set and the
- * checks can start. More candidates can be added afterwards
- */
-
-static void
-fs_rtp_stream_remote_candidates_added (FsStream *stream)
-{
-  FsRtpStream *self = FS_RTP_STREAM (stream);
-
-  fs_stream_transmitter_remote_candidates_added (
-      self->priv->stream_transmitter);
+  g_object_unref (st);
+  return ret;
 }
 
 /**
- * fs_rtp_stream_select_candidate_pair:
- * @stream: a #FsStream
- * @lfoundation: The foundation of the local candidate to be selected
- * @rfoundation: The foundation of the remote candidate to be selected
- * @error: location of a #GError, or NULL if no error occured
+ * fs_rtp_stream_force_remote_candidates
  *
- * This function selects one pair of candidates to be selected to start
- * sending media on.
- *
- * Returns: TRUE if the candidate pair could be selected, FALSE otherwise
+ * Implement FsStream -> force_remote_candidates
+ * by calling the same function in the stream transmittrer
  */
 
 static gboolean
-fs_rtp_stream_select_candidate_pair (FsStream *stream, gchar *lfoundation,
-                                     gchar *rfoundation, GError **error)
+fs_rtp_stream_force_remote_candidates (FsStream *stream,
+    GList *remote_candidates,
+    GError **error)
 {
   FsRtpStream *self = FS_RTP_STREAM (stream);
+  FsStreamTransmitter *st = fs_rtp_stream_get_stream_transmitter (self, error);
+  gboolean ret = FALSE;
 
-  return fs_stream_transmitter_select_candidate_pair (
-      self->priv->stream_transmitter, lfoundation, rfoundation, error);
+  if (!st)
+    return FALSE;
+
+
+  ret = fs_stream_transmitter_force_remote_candidates (
+      self->priv->stream_transmitter, remote_candidates,
+      error);
+
+  g_object_unref (st);
+  return ret;
 }
 
 
@@ -481,14 +575,18 @@ fs_rtp_stream_set_remote_codecs (FsStream *stream,
   FsRtpStream *self = FS_RTP_STREAM (stream);
   GList *item = NULL;
   FsMediaType media_type;
+  FsRtpSession *session = fs_rtp_stream_get_session (self, error);
+
+  if (!session)
+    return FALSE;
 
   if (remote_codecs == NULL) {
     g_set_error (error, FS_ERROR, FS_ERROR_INVALID_ARGUMENTS,
       "You can not set NULL remote codecs");
-    return FALSE;
+    goto error;
   }
 
-  g_object_get (self->priv->session, "media-type", &media_type, NULL);
+  g_object_get (session, "media-type", &media_type, NULL);
 
   for (item = g_list_first (remote_codecs); item; item = g_list_next (item))
   {
@@ -498,39 +596,49 @@ fs_rtp_stream_set_remote_codecs (FsStream *stream,
     {
       g_set_error (error, FS_ERROR, FS_ERROR_INVALID_ARGUMENTS,
           "The codec must have an encoding name");
-      return FALSE;
+      goto error;
     }
     if (codec->id < 0 || codec->id > 128)
     {
       g_set_error (error, FS_ERROR, FS_ERROR_INVALID_ARGUMENTS,
           "The codec id must be between 0 ans 128 for %s",
           codec->encoding_name);
-      return FALSE;
+      goto error;
     }
     if (codec->clock_rate == 0)
     {
       g_set_error (error, FS_ERROR, FS_ERROR_INVALID_ARGUMENTS,
-          "The codec %s must has a non-0 clock rate", codec->encoding_name);
-      return FALSE;
+          "The codec %s must have a non-0 clock rate", codec->encoding_name);
+      goto error;
     }
     if (codec->media_type != media_type)
     {
       g_set_error (error, FS_ERROR, FS_ERROR_INVALID_ARGUMENTS,
           "The media type for codec %s is not %s", codec->encoding_name,
           fs_media_type_to_string (media_type));
-      return FALSE;
+      goto error;
     }
   }
 
-  if (fs_rtp_session_negotiate_codecs (self->priv->session, remote_codecs,
-          stream, error)) {
+  if (self->priv->new_remote_codecs_cb (self, remote_codecs, error,
+          self->priv->user_data_for_cb))
+  {
+    FS_RTP_SESSION_LOCK (session);
     if (self->remote_codecs)
       fs_codec_list_destroy (self->remote_codecs);
     self->remote_codecs = fs_codec_list_copy (remote_codecs);
-    return TRUE;
+    FS_RTP_SESSION_UNLOCK (session);
   } else {
-    return FALSE;
+    goto error;
   }
+
+  g_object_unref (session);
+  return TRUE;
+
+ error:
+
+  g_object_unref (session);
+  return FALSE;
 }
 
 /**
@@ -540,7 +648,12 @@ fs_rtp_stream_set_remote_codecs (FsStream *stream,
  * @direction: the initial #FsDirection for this stream
  * @stream_transmitter: the #FsStreamTransmitter for this stream, one
  *   reference to it will be eaten
- *
+ * @new_remote_codecs_cb: Callback called when the remote codecs change
+ * (ie when fs_rtp_stream_set_remote_codecs() is called). One must hold
+ * the session lock across calls.
+ * @known_source_packet_received: Callback called when a packet from a
+ * known source is receive.
+ * @user_data: User data for the callbacks.
  * This function create a new stream
  *
  * Returns: the newly created string or NULL on error
@@ -548,18 +661,33 @@ fs_rtp_stream_set_remote_codecs (FsStream *stream,
 
 FsRtpStream *
 fs_rtp_stream_new (FsRtpSession *session,
-                   FsRtpParticipant *participant,
-                   FsStreamDirection direction,
-                   FsStreamTransmitter *stream_transmitter,
-                   GError **error)
+    FsRtpParticipant *participant,
+    FsStreamDirection direction,
+    FsStreamTransmitter *stream_transmitter,
+    stream_new_remote_codecs_cb new_remote_codecs_cb,
+    stream_known_source_packet_receive_cb known_source_packet_received_cb,
+    gpointer user_data_for_cb,
+    GError **error)
 {
-  FsRtpStream *self = g_object_new (FS_TYPE_RTP_STREAM,
+  FsRtpStream *self;
+
+  g_return_val_if_fail (session, NULL);
+  g_return_val_if_fail (participant, NULL);
+  g_return_val_if_fail (stream_transmitter, NULL);
+  g_return_val_if_fail (new_remote_codecs_cb, NULL);
+  g_return_val_if_fail (known_source_packet_received_cb, NULL);
+
+  self = g_object_new (FS_TYPE_RTP_STREAM,
     "session", session,
     "participant", participant,
     "direction", direction,
     "stream-transmitter", stream_transmitter,
     NULL);
 
+  self->priv->new_remote_codecs_cb = new_remote_codecs_cb;
+  self->priv->known_source_packet_received_cb = known_source_packet_received_cb;
+  self->priv->user_data_for_cb = user_data_for_cb;
+
   if (self->priv->construction_error) {
     g_propagate_error (error, self->priv->construction_error);
     g_object_unref (self);
@@ -575,22 +703,51 @@ _local_candidates_prepared (FsStreamTransmitter *stream_transmitter,
     gpointer user_data)
 {
   FsRtpStream *self = FS_RTP_STREAM (user_data);
+  GstElement *conf = NULL;
+  FsRtpSession *session = fs_rtp_stream_get_session (self, NULL);
 
-  g_signal_emit_by_name (self, "local-candidates-prepared");
+  if (!session)
+    return;
+
+  g_object_get (session, "conference", &conf, NULL);
+
+  gst_element_post_message (conf,
+      gst_message_new_element (GST_OBJECT (conf),
+          gst_structure_new ("farsight-local-candidates-prepared",
+              "stream", FS_TYPE_STREAM, self,
+              NULL)));
+
+  gst_object_unref (conf);
+  g_object_unref (session);
 }
 
 
 static void
 _new_active_candidate_pair (
     FsStreamTransmitter *stream_transmitter,
-    FsCandidate *candidate1,
-    FsCandidate *candidate2,
+    FsCandidate *local_candidate,
+    FsCandidate *remote_candidate,
     gpointer user_data)
 {
   FsRtpStream *self = FS_RTP_STREAM (user_data);
+  FsRtpSession *session = fs_rtp_stream_get_session (self, NULL);
+  GstElement *conf = NULL;
 
-  g_signal_emit_by_name (self, "new-active-candidate-pair",
-    candidate1, candidate2);
+  if (!session)
+    return;
+
+  g_object_get (session, "conference", &conf, NULL);
+
+  gst_element_post_message (conf,
+      gst_message_new_element (GST_OBJECT (conf),
+          gst_structure_new ("farsight-new-active-candidate-pair",
+              "stream", FS_TYPE_STREAM, self,
+              "local-candidate", FS_TYPE_CANDIDATE, local_candidate,
+              "remote-candidate", FS_TYPE_CANDIDATE, remote_candidate,
+              NULL)));
+
+  gst_object_unref (conf);
+  g_object_unref (session);
 }
 
 
@@ -601,8 +758,23 @@ _new_local_candidate (
     gpointer user_data)
 {
   FsRtpStream *self = FS_RTP_STREAM (user_data);
+  FsRtpSession *session = fs_rtp_stream_get_session (self, NULL);
+  GstElement *conf = NULL;
 
-  g_signal_emit_by_name (self, "new-local-candidate", candidate);
+  if (!session)
+    return;
+
+  g_object_get (session, "conference", &conf, NULL);
+
+  gst_element_post_message (conf,
+      gst_message_new_element (GST_OBJECT (conf),
+          gst_structure_new ("farsight-new-local-candidate",
+              "stream", FS_TYPE_STREAM, self,
+              "candidate", FS_TYPE_CANDIDATE, candidate,
+              NULL)));
+
+  gst_object_unref (conf);
+  g_object_unref (session);
 }
 
 static void
@@ -618,130 +790,150 @@ _transmitter_error (
   fs_stream_emit_error (stream, errorno, error_msg, debug_msg);
 }
 
+static void
+_known_source_packet_received (FsStreamTransmitter *st,
+    guint component,
+    GstBuffer *buffer,
+    FsRtpStream *self)
+{
+  self->priv->known_source_packet_received_cb (self, component, buffer,
+      self->priv->user_data_for_cb);
+}
+
+static void
+_state_changed (FsStreamTransmitter *stream_transmitter,
+    guint component,
+    FsStreamState state,
+    gpointer user_data)
+{
+  FsRtpStream *self = FS_RTP_STREAM (user_data);
+  FsRtpSession *session = fs_rtp_stream_get_session (self, NULL);
+  GstElement *conf = NULL;
+
+  if (!session)
+    return;
+
+  g_object_get (session, "conference", &conf, NULL);
+
+  gst_element_post_message (conf,
+      gst_message_new_element (GST_OBJECT (conf),
+          gst_structure_new ("farsight-component-state-changed",
+              "stream", FS_TYPE_STREAM, self,
+              "component", G_TYPE_UINT, component,
+              "state", FS_TYPE_STREAM_STATE, state,
+              NULL)));
+
+  gst_object_unref (conf);
+  g_object_unref (session);
+
+  if (component == 1 && state == FS_STREAM_STATE_FAILED)
+    fs_stream_emit_error (FS_STREAM (self), FS_ERROR_CONNECTION_FAILED,
+        "Could not establish connection", "Could not establish connection"
+        " on the RTP component");
+}
+
+static void
+_substream_src_pad_added (FsRtpSubStream *substream, GstPad *pad,
+                          FsCodec *codec, gpointer user_data)
+{
+  FsStream *stream = FS_STREAM (user_data);
+
+  fs_stream_emit_src_pad_added (stream, pad, codec);
+}
+
+static void
+_substream_error (FsRtpSubStream *substream,
+    gint errorno,
+    gchar *error_msg,
+    gchar *debug_msg,
+    gpointer user_data)
+{
+  FsStream *stream = FS_STREAM (user_data);
+
+  fs_stream_emit_error (stream, errorno, error_msg, debug_msg);
+}
 
 /**
- * fs_rtp_stream_add_substream:
+ * fs_rtp_stream_add_substream_unlock:
  * @stream: a #FsRtpStream
  * @substream: the #FsRtpSubStream to associate with this stream
  *
  * This functions associates a substream with this stream
  *
+ * You must enter this function with the session lock held and it will release
+ * it.
+ *
  * Returns: TRUE on success, FALSE on failure
  */
 gboolean
-fs_rtp_stream_add_substream (FsRtpStream *stream,
+fs_rtp_stream_add_substream_unlock (FsRtpStream *stream,
     FsRtpSubStream *substream,
     GError **error)
 {
-  FsCodec *codec = NULL;
   gboolean ret = TRUE;
+  FsRtpSession *session = fs_rtp_stream_get_session (stream, error);
 
-  FS_RTP_SESSION_LOCK (stream->priv->session);
-  stream->priv->substreams = g_list_prepend (stream->priv->substreams,
+  if (!session)
+    return FALSE;
+
+  stream->substreams = g_list_prepend (stream->substreams,
       substream);
   g_object_set (substream,
       "stream", stream,
       "receiving", ((stream->priv->direction & FS_DIRECTION_RECV) != 0),
       NULL);
 
-  g_object_get (substream, "codec", &codec, NULL);
+  g_signal_connect (substream, "src-pad-added",
+                    G_CALLBACK (_substream_src_pad_added), stream);
+  g_signal_connect (substream, "codec-changed",
+                    G_CALLBACK (_substream_codec_changed), stream);
+  g_signal_connect (substream, "error",
+                    G_CALLBACK (_substream_error), stream);
 
   /* Only announce a pad if it has a codec attached to it */
-  if (codec) {
-    ret = fs_rtp_sub_stream_add_output_ghostpad_locked (substream, error);
-    fs_codec_destroy (codec);
+  if (substream->codec)
+    ret = fs_rtp_sub_stream_add_output_ghostpad_unlock (substream, error);
+  else
+    FS_RTP_SESSION_UNLOCK (session);
 
-    fs_rtp_stream_maybe_emit_codecs_changed (stream, substream);
-  }
-
-  FS_RTP_SESSION_UNLOCK (stream->priv->session);
+  g_object_unref (session);
 
   return ret;
 }
 
-gboolean
-fs_rtp_stream_knows_ssrc_locked (FsRtpStream *stream, guint32 ssrc)
-{
-  GList *elem;
-
-  for (elem = g_list_first (stream->priv->known_ssrcs);
-       elem;
-       elem = g_list_next (elem))
-  {
-    guint32 tmp_ssrc = GPOINTER_TO_UINT (elem->data);
-    if (tmp_ssrc == ssrc)
-      return TRUE;
-  }
-
-  return FALSE;
-}
-
 /**
- * fs_rtp_stream_invalidate_codec_locked:
- * @stream: A #FsRtpStream
- * @pt: The payload type to invalidate (does nothing if it does not match)
- * @codec: The new fscodec (the substream is invalidated if it not using this
- *  codec). You can pass NULL to match any codec.
- *
- * This function will start the process that invalidates the codec
- * for this rtpbin, if it doesnt match the passed codec
- *
- * You must hold the session lock to call it.
- */
-
-void
-fs_rtp_stream_invalidate_codec_locked (FsRtpStream *stream,
-    gint pt,
-    const FsCodec *codec)
-{
-  GList *item;
-
-  for (item = g_list_first (stream->priv->substreams);
-       item;
-       item = g_list_next (item))
-    fs_rtp_sub_stream_invalidate_codec_locked (item->data, pt, codec);
-}
-
-
-
-static gboolean
-_idle_emit_recv_codecs_changed (gpointer data)
-{
-  FsRtpStream *stream = FS_RTP_STREAM (data);
-
-  g_signal_emit_by_name (stream, "recv-codecs-changed");
-
-  FS_RTP_SESSION_LOCK (stream->priv->session);
-  stream->priv->recv_codecs_changed_idle_id = 0;
-  FS_RTP_SESSION_UNLOCK (stream->priv->session);
-  return FALSE;
-}
-
-/**
- * fs_stream_maybe_emit_codecs_changed:
- * @stream: a #FsRtpStream
+ *  _substream_codec_changed
  * @substream: The #FsRtpSubStream that may have a new receive codec
+ * @stream: a #FsRtpStream
  *
- * This function checks if the specified substream introduces not a new codec
- * not present in another substream and if it does, it schedules an idle task
- * to emit the signal on the main thread.
+ * This function checks if the specified substream introduces a new codec
+ * not present in another substream and if it does, it emits a GstMessage
+ * and the notify signal
  */
 
-void
-fs_rtp_stream_maybe_emit_codecs_changed (FsRtpStream *stream,
-    FsRtpSubStream *substream)
+static void
+_substream_codec_changed (FsRtpSubStream *substream,
+    FsRtpStream *stream)
 {
   GList *substream_item = NULL;
-  FsCodec *codec = NULL;
+  GList *codeclist = NULL;
+  FsRtpSession *session = fs_rtp_stream_get_session (stream, NULL);
 
-  g_object_get (substream, "codec", &codec, NULL);
-
-  if (!codec)
+  if (!session)
     return;
 
-  FS_RTP_SESSION_LOCK (stream->priv->session);
+  FS_RTP_SESSION_LOCK (session);
 
-  for (substream_item = stream->priv->substreams;
+  if (!substream->codec)
+  {
+    FS_RTP_SESSION_UNLOCK (session);
+    g_object_unref (session);
+    return;
+  }
+
+  codeclist = g_list_prepend (NULL, fs_codec_copy (substream->codec));
+
+  for (substream_item = stream->substreams;
        substream_item;
        substream_item = g_list_next (substream_item))
   {
@@ -749,64 +941,78 @@ fs_rtp_stream_maybe_emit_codecs_changed (FsRtpStream *stream,
 
     if (othersubstream != substream)
     {
-      FsCodec *othercodec = NULL;
-
-      g_object_get (othersubstream, "codec", &othercodec, NULL);
-
-      if (othercodec && ! fs_codec_are_equal (codec, othercodec))
+      if (othersubstream->codec)
       {
-        fs_codec_destroy (othercodec);
-        break;
-      }
+        if (fs_codec_are_equal (substream->codec, othersubstream->codec))
+          break;
 
-      fs_codec_destroy (othercodec);
+        if (!_codec_list_has_codec (codeclist, othersubstream->codec))
+          codeclist = g_list_append (codeclist,
+              fs_codec_copy (othersubstream->codec));
+      }
     }
   }
 
+  FS_RTP_SESSION_UNLOCK (session);
+
   if (substream_item == NULL)
-    if (!stream->priv->recv_codecs_changed_idle_id)
-      stream->priv->recv_codecs_changed_idle_id =
-        g_idle_add (_idle_emit_recv_codecs_changed, stream);
-
-  FS_RTP_SESSION_UNLOCK (stream->priv->session);
-
-  fs_codec_destroy (codec);
-}
-
-/**
- * fs_rtp_stream_add_known_ssrc:
- * @stream: a #FsRtpStream
- * @ssrc: the SSRC to add
- *
- * Adds a SSRC to the list of known SSRCs for this stream
- */
-
-void
-fs_rtp_stream_add_known_ssrc (FsRtpStream *stream,
-    guint32 ssrc)
-{
-  FS_RTP_SESSION_LOCK (stream->priv->session);
-  if (!fs_rtp_stream_knows_ssrc_locked (stream, ssrc))
   {
-    stream->priv->known_ssrcs = g_list_prepend (stream->priv->known_ssrcs,
-        GUINT_TO_POINTER (ssrc));
+    GstElement *conf = NULL;
+
+    g_object_notify (G_OBJECT (stream), "current-recv-codecs");
+
+    g_object_get (session, "conference", &conf, NULL);
+
+    gst_element_post_message (conf,
+        gst_message_new_element (GST_OBJECT (conf),
+            gst_structure_new ("farsight-recv-codecs-changed",
+                "stream", FS_TYPE_STREAM, stream,
+                "codecs", FS_TYPE_CODEC_LIST, codeclist,
+                NULL)));
+
+    gst_object_unref (conf);
   }
-  FS_RTP_SESSION_UNLOCK (stream->priv->session);
+
+  fs_codec_list_destroy (codeclist);
+  g_object_unref (session);
 }
 
 /**
- * fs_rtp_stream_remove_known_ssrc:
+ * fs_rtp_stream_set_negotiated_codecs_unlock
  * @stream: a #FsRtpStream
- * @ssrc: the SSRC to remove
+ * @codecs: The #GList of #FsCodec to set for the negotiated-codecs property
  *
- * Removes the ssrc from the list of known ssrcs
+ * This function sets the value of the FsStream:negotiated-codecs property.
+ * Unlike most other functions in this element, it TAKES the reference to the
+ * codecs, so you have to give it its own copy.
+ *
+ * You must enter this function with the session lock held and it will release
+ * it.
  */
 void
-fs_rtp_stream_remove_known_ssrc (FsRtpStream *stream,
-    guint32 ssrc)
+fs_rtp_stream_set_negotiated_codecs_unlock (FsRtpStream *stream,
+    GList *codecs)
 {
-  FS_RTP_SESSION_LOCK (stream->priv->session);
-  stream->priv->known_ssrcs = g_list_remove_all (stream->priv->known_ssrcs,
-      GUINT_TO_POINTER (ssrc));
-  FS_RTP_SESSION_UNLOCK (stream->priv->session);
+  FsRtpSession *session = fs_rtp_stream_get_session (stream, NULL);
+
+  if (!session)
+    return;
+
+  if (fs_codec_list_are_equal (stream->negotiated_codecs, codecs))
+  {
+    fs_codec_list_destroy (codecs);
+    FS_RTP_SESSION_UNLOCK (session);
+    return;
+  }
+
+  if (stream->negotiated_codecs)
+    fs_codec_list_destroy (stream->negotiated_codecs);
+
+  stream->negotiated_codecs = codecs;
+
+  FS_RTP_SESSION_UNLOCK (session);
+
+  g_object_notify (G_OBJECT (stream), "negotiated-codecs");
+
+  g_object_unref (session);
 }
