@@ -30,8 +30,18 @@ GstElement *setup_pipeline (FsTransmitter *trans, GCallback cb);
 void setup_fakesrc (FsTransmitter *trans, GstElement *pipeline,
   guint component_id);
 
-void _stream_transmitter_error (FsStreamTransmitter *streamtransmitter,
+void stream_transmitter_error (FsStreamTransmitter *streamtransmitter,
   gint errorno, gchar *error_msg, gchar *debug_msg, gpointer user_data);
 
+gboolean bus_error_callback (GstBus *bus, GstMessage *message,
+    gpointer user_data);
+
+void test_transmitter_creation (gchar *transmitter_name);
+
+extern GPid stund_pid;
+
+void setup_stund (void);
+void teardown_stund (void);
+
 
 #endif /* __GENERIC_H__ */
