@@ -25,6 +25,8 @@ import socket
 import threading
 import weakref
 
+import signal
+
 try:
     import pygtk
     pygtk.require("2.0")
@@ -94,9 +96,17 @@ def make_video_sink(pipeline, xid, name, async=True):
 
 class FsUIPipeline:
     "Object to wrap the GstPipeline"
+
+    def int_handler(self, sig, frame):
+        try:
+            self.pipeline.to_dot_file(0, "pipelinedump")
+        except:
+            pass
+        sys.exit(2)
     
     def __init__(self, elementname="fsrtpconference"):
         self.pipeline = gst.Pipeline()
+        signal.signal(signal.SIGINT, self.int_handler)
         notifier = farsight.ElementAddedNotifier()
         notifier.connect("element-added", self.element_added_cb)
         notifier.add(self.pipeline)
@@ -112,6 +122,7 @@ class FsUIPipeline:
         if AUDIO:
             self.audiosource = FsUIAudioSource(self.pipeline)
             self.audiosession = FsUISession(self.conf, self.audiosource)
+            self.adder = None
         self.pipeline.set_state(gst.STATE_PLAYING)
 
     def __del__(self):
@@ -135,15 +146,46 @@ class FsUIPipeline:
 
     def async_handler(self, bus, message):
         "Async handler to print messages"
-        if message.type != gst.MESSAGE_STATE_CHANGED \
-               and message.type != gst.MESSAGE_ASYNC_DONE:
-            print message.type
         if message.type == gst.MESSAGE_ERROR:
             print message.src.get_name(), ": ", message.parse_error()
         elif message.type == gst.MESSAGE_WARNING:
             print message.src.get_name(), ": ", message.parse_warning()
         elif message.type == gst.MESSAGE_ELEMENT:
-            print message.src.get_name(), ": ", message.structure.get_name()
+            if message.structure.has_name("dtmf-event"):
+                print "dtmf-event: %d" % message.structure["number"]
+            elif message.structure.has_name("farsight-local-candidates-prepared"):
+                message.structure["stream"].uistream.local_candidates_prepared()
+
+            elif message.structure.has_name("farsight-new-local-candidate"):
+                message.structure["stream"].uistream.new_local_candidate(
+                    message.structure["candidate"])
+            elif message.structure.has_name("farsight-codecs-changed"):
+                print message.src.get_name(), ": ", message.structure.get_name()
+                message.structure["session"].uisession.codecs_changed()
+                if AUDIO and message.structure["session"] == self.audiosession.fssession:
+                    self.codecs_changed_audio()
+                if VIDEO and  message.structure["session"] == self.videosession.fssession:
+                    self.codecs_changed_video()
+            elif message.structure.has_name("farsight-send-codec-changed"):
+                print message.src.get_name(), ": ", message.structure.get_name()
+                print "send codec changed: " + message.structure["codec"].to_string()
+                if AUDIO and message.structure["session"] == self.audiosession.fssession:
+                    self.codecs_changed_audio()
+                if VIDEO and message.structure["session"] == self.videosession.fssession:
+                    self.codecs_changed_video()
+            elif message.structure.has_name("farsight-recv-codecs-changed"):
+                print message.src.get_name(), ": ", message.structure.get_name()
+                message.structure["stream"].uistream.recv_codecs_changed( \
+                    message.structure["codecs"])
+                
+                
+            elif message.structure.has_name("farsight-error"):
+                print "Async error ("+ str(message.structure["error-no"]) +"): " + message.structure["error-msg"] +" --- "+ message.structure["debug-msg"]
+            else:
+                print message.src.get_name(), ": ", message.structure.get_name()
+        elif message.type != gst.MESSAGE_STATE_CHANGED \
+                 and message.type != gst.MESSAGE_ASYNC_DONE:
+            print message.type
         
         return True
 
@@ -172,15 +214,41 @@ class FsUIPipeline:
     def link_audio_sink(self, pad):
         "Link the audio sink to the pad"
         print >>sys.stderr, "LINKING AUDIO SINK"
-        self.audiosink = gst.element_factory_make("alsasink")
-        self.pipeline.add(self.audiosink)
-        self.audiosink.set_state(gst.STATE_PLAYING)
-        pad.link(self.audiosink.get_pad("sink"))
+        if not self.adder:
+            audiosink = gst.element_factory_make("alsasink")
+            audiosink.set_property("buffer-time", 50000)
+            self.pipeline.add(audiosink)
+
+            try:
+                self.adder = gst.element_factory_make("liveadder")
+            except gst.ElementNotFoundError:
+                audiosink.set_state(gst.STATE_PLAYING)
+                pad.link(audiosink.get_pad("sink"))
+                return
+            self.pipeline.add(self.adder)
+            audiosink.set_state(gst.STATE_PLAYING)
+            self.adder.link(audiosink)
+            self.adder.set_state(gst.STATE_PLAYING)
+        convert1 = gst.element_factory_make("audioconvert")
+        self.pipeline.add(convert1)
+        resample = gst.element_factory_make("audioresample")
+        self.pipeline.add(resample)
+        convert2 = gst.element_factory_make("audioconvert")
+        self.pipeline.add(convert2)
+        convert1.link(resample)
+        resample.link(convert2)
+        convert2.link(self.adder)
+        pad.link(convert1.get_pad("sink"))
+        convert2.set_state(gst.STATE_PLAYING)
+        resample.set_state(gst.STATE_PLAYING)
+        convert1.set_state(gst.STATE_PLAYING)
 
     def element_added_cb(self, notifier, bin, element):
         if element.get_factory().get_name() == "x264enc":
             element.set_property("byte-stream", True)
             element.set_property("bitrate", 128)
+        elif element.get_factory().get_name() == "gstrtpbin":
+            element.set_property("latency", 100)
             
 
 class FsUISource:
@@ -216,8 +284,8 @@ class FsUISource:
     def get_src_pad(self, name="src%d"):
         "Gets a source pad from the source"
         queue = gst.element_factory_make("queue")
-        queue.set_properties("leaky", 2,
-                             "max-size-time", 50*gst.MSECOND)
+        queue.set_property("leaky", 2)
+        queue.set_property("max-size-time", 50*gst.MSECOND)
         requestpad = self.tee.get_request_pad(name)
         self.pipeline.add(queue)
         requestpad.link(queue.get_static_pad("sink"))
@@ -243,11 +311,16 @@ class FsUIVideoSource(FsUISource):
         if CAMERA:
             source = gst.element_factory_make("v4l2src")
             source.set_property("device", CAMERA)
+            bin.add(source)
         else:
             source = gst.element_factory_make("videotestsrc")
             source.set_property("is-live", 1)
-            
-        bin.add(source)
+            bin.add(source)
+            overlay = gst.element_factory_make("timeoverlay")
+            overlay.set_property("font-desc", "Sans 32")
+            bin.add(overlay)
+            source.link(overlay)
+            source=overlay
 
         filter = gst.element_factory_make("capsfilter")
         filter.set_property("caps", gst.Caps("video/x-raw-yuv , width=[300,500] , height=[200,500], framerate=[20/1,30/1]"))
@@ -270,11 +343,12 @@ class FsUIAudioSource(FsUISource):
         return farsight.MEDIA_TYPE_AUDIO
 
     def make_source(self):
-        #source = gst.element_factory_make("audiotestsrc")
-        #source.set_property("is-live", True)
-        #source.set_property("wave", 5)
-        #return source
-        return gst.element_factory_make("alsasrc")
+        source = gst.element_factory_make("audiotestsrc")
+        source.set_property("is-live", True)
+        source.set_property("wave", 5)
+        return source
+        #return gst.element_factory_make("alsasrc")
+        #return gst.element_factory_make("gconfaudiosrc")
 
 
 
@@ -284,53 +358,62 @@ class FsUISession:
     def __init__(self, conference, source):
         self.conference = conference
         self.source = source
-        self.streams = weakref.WeakValueDictionary()
+        self.streams = []
         self.fssession = conference.new_session(source.get_type())
+        self.fssession.uisession = self
         if source.get_type() == farsight.MEDIA_TYPE_VIDEO:
             # We prefer H263-1998 because we know it works
             # We don't know if the others do work
             # We know H264 doesn't work for now or anything else
             # that needs to send config data
-            self.fssession.set_property("local-codecs-config",
-                                        [farsight.Codec(farsight.CODEC_ID_ANY,
-                                                      "H264",
-                                                      farsight.MEDIA_TYPE_VIDEO,
-                                                      0),
-                                        farsight.Codec(farsight.CODEC_ID_ANY,
-                                                     "H263-1998",
-                                                     farsight.MEDIA_TYPE_VIDEO,
-                                                     0),
-                                        farsight.Codec(farsight.CODEC_ID_ANY,
-                                                     "H263",
-                                                     farsight.MEDIA_TYPE_VIDEO,
-                                                     0)])
+            self.fssession.set_codec_preferences( [ \
+                farsight.Codec(farsight.CODEC_ID_ANY,
+                               "THEORA",
+                               farsight.MEDIA_TYPE_VIDEO,
+                               90000),
+                farsight.Codec(farsight.CODEC_ID_ANY,
+                               "H264",
+                               farsight.MEDIA_TYPE_VIDEO,
+                               0),
+                farsight.Codec(farsight.CODEC_ID_ANY,
+                               "H263-1998",
+                               farsight.MEDIA_TYPE_VIDEO,
+                               0),
+                farsight.Codec(farsight.CODEC_ID_ANY,
+                               "H263",
+                               farsight.MEDIA_TYPE_VIDEO,
+                               0)
+                ])
         elif source.get_type() == farsight.MEDIA_TYPE_AUDIO:
-            self.fssession.set_property("local-codecs-config",
-                                      [farsight.Codec(farsight.CODEC_ID_ANY,
-                                                      "PCMA",
-                                                      farsight.MEDIA_TYPE_AUDIO,
-                                                      0),
-                                       farsight.Codec(farsight.CODEC_ID_ANY,
-                                                      "PCMU",
-                                                      farsight.MEDIA_TYPE_AUDIO,
-                                                      0)])
-        self.fssession.connect("new-negotiated-codecs",
-                             self.__new_negotiated_codecs)
+            self.fssession.set_codec_preferences( [ \
+                farsight.Codec(farsight.CODEC_ID_ANY,
+                               "PCMA",
+                               farsight.MEDIA_TYPE_AUDIO,
+                               0),
+                farsight.Codec(farsight.CODEC_ID_ANY,
+                               "PCMU",
+                               farsight.MEDIA_TYPE_AUDIO,
+                               0),
+                # The gst speexenc element breaks timestamps
+                farsight.Codec(farsight.CODEC_ID_DISABLE,
+                               "SPEEX",
+                               farsight.MEDIA_TYPE_AUDIO,
+                               16000),
+                # Sadly, vorbis is not currently compatible with live streaming :-(
+                farsight.Codec(farsight.CODEC_ID_DISABLE,
+                               "VORBIS",
+                               farsight.MEDIA_TYPE_AUDIO,
+                               0),
+                ])
+
         self.sourcepad = self.source.get_src_pad()
         self.sourcepad.link(self.fssession.get_property("sink-pad"))
 
     def __del__(self):
         self.sourcepad(unlink)
         self.source.put_src_pad(self.sourcepad)
-        
-    def __new_negotiated_codecs(self, session):
-        "Callback from FsSession"
-        for s in self.streams.valuerefs():
-            try:
-                s().new_negotiated_codecs()
-            except AttributeError:
-                pass
-            
+    def __stream_finalized(self, s):
+        self.streams.remove(s)
             
     def new_stream(self, id, participant):
         "Creates a new stream for a specific participant"
@@ -347,9 +430,37 @@ class FsUISession:
                                              farsight.DIRECTION_BOTH,
                                              TRANSMITTER, transmitter_params)
         stream = FsUIStream(id, self, participant, realstream)
-        self.streams[id] = stream
+        self.streams.append(weakref.ref(stream, self.__stream_finalized))
         return stream
-    
+
+    def dtmf_start(self, event, method):
+        if (event == "*"):
+            event = farsight.DTMF_EVENT_STAR
+        elif (event == "#"):
+            event = farsight.DTMF_EVENT_POUND
+        else:
+            event = int(event)
+        self.fssession.start_telephony_event(event, 2, method)
+        
+    def dtmf_stop(self, method):
+        self.fssession.stop_telephony_event(method)
+
+    def codecs_changed(self):
+        "Callback from FsSession"
+        for s in self.streams:
+            try:
+                s().codecs_changed()
+            except AttributeError:
+                pass
+
+    def send_stream_codecs(self, codecs, sourcestream):
+        for s in self.streams:
+            stream = s()
+            if stream and stream is not sourcestream:
+                stream.connect.send_codecs(stream.participant.id,
+                                           sourcestream.id,
+                                           codecs,
+                                           sourcestream.participant.id)
 
 class FsUIStream:
     "One participant in one session"
@@ -360,17 +471,17 @@ class FsUIStream:
         self.participant = participant
         self.fsstream = fsstream
         self.connect = participant.connect
-        self.fsstream.connect("local-candidates-prepared",
-                            self.__local_candidates_prepared)
-        self.fsstream.connect("new-local-candidate",
-                            self.__new_local_candidate)
+        self.fsstream.uistream = self
         self.fsstream.connect("src-pad-added", self.__src_pad_added)
-        self.newcodecs = []
+        self.send_codecs = False
+        self.last_codecs = None
+        self.last_stream_codecs = None
+        self.candidates = []
 
-    def __local_candidates_prepared(self, stream):
+    def local_candidates_prepared(self):
         "Callback from FsStream"
         self.connect.send_candidates_done(self.participant.id, self.id)
-    def __new_local_candidate(self, stream, candidate):
+    def new_local_candidate(self, candidate):
         "Callback from FsStream"
         self.connect.send_candidate(self.participant.id, self.id, candidate)
     def __src_pad_added(self, stream, pad, codec):
@@ -382,47 +493,77 @@ class FsUIStream:
 
     def candidate(self, candidate):
         "Callback for the network object."
-        self.fsstream.add_remote_candidate(candidate)
+        self.candidates.append(candidate)
     def candidates_done(self):
         "Callback for the network object."
-        self.fsstream.remote_candidates_added()
-    def codec(self, codec):
-        "Callback for the network object. Stores the codec"
-        
-        self.newcodecs.append(codec)
-        
-    def codecs_done(self):
-        """Callback for the network object.
+        self.fsstream.set_remote_candidates(self.candidates)
+        self.candidates = []
+    def codecs(self, codecs):
+        "Callback for the network object. Set the codecs"
 
-        When all the codecs have been received, we can set them on the stream.
-        """
-        if len(self.newcodecs) > 0:
-            self.codecs = self.newcodecs
-            self.newcodecs = []
+        print "Remote codecs"
+        for c in codecs:
+            print "Got remote codec from %s/%s %s" % \
+                  (self.participant.id, self.id, c.to_string())
+        oldcodecs = self.fsstream.get_property("remote-codecs")
+        if oldcodecs == codecs:
+            return
         try:
-            self.fsstream.set_remote_codecs(self.codecs)
+            self.fsstream.set_remote_codecs(codecs)
         except AttributeError:
             print "Tried to set codecs with 0 codec"
-            
+        self.send_local_codecs()
+        self.send_stream_codecs()
+
+
     def send_local_codecs(self):
         "Callback for the network object."
-        codecs = self.session.fssession.get_property("negotiated-codecs")
-        if codecs is None or len(codecs) == 0:
-            codecs = self.session.fssession.get_property("local-codecs")
-        for codec in codecs:
-            self.connect.send_codec(self.participant.id, self.id, codec)
-        self.connect.send_codecs_done(self.participant.id, self.id)
+        self.send_codecs = True
+        self.check_send_local_codecs()
 
-    def new_negotiated_codecs(self):
-        """Callback for the network object.
+    def codecs_changed(self):
+        self.check_send_local_codecs()
+        self.send_stream_codecs()
 
-        We only send our negotiated codecs to the server (participant 1) or
-        to everyone if we are the server.
-        """
-        if self.participant.id == 1 or self.connect.myid == 1:
-            for codec in self.session.fssession.get_property("negotiated-codecs"):
-                self.connect.send_codec(self.participant.id, self.id, codec)
-            self.connect.send_codecs_done(self.participant.id, self.id)
+    def check_send_local_codecs(self):
+        "Internal function to send our local codecs when they're ready"
+        if not self.send_codecs:
+            return
+        if not self.session.fssession.get_property("codecs-ready"):
+            print "Codecs are not ready"
+            return
+        codecs = self.session.fssession.get_property("codecs")
+        assert(codecs is not None and len(codecs) > 0)
+        if (codecs == self.last_codecs):
+            return
+        self.last_codecs = codecs
+        print "sending local codecs"
+        self.connect.send_codecs(self.participant.id, self.id, codecs)
+
+    def send_stream_codecs(self):
+        if not self.connect.is_server:
+            return
+        if not self.session.fssession.get_property("codecs-ready"):
+            return
+        codecs = self.fsstream.get_property("negotiated-codecs")
+        if codecs:
+            self.session.send_stream_codecs(codecs, self)
+
+    def recv_codecs_changed(self, codecs):
+        self.participant.recv_codecs_changed()
+
+
+    def __remove_from_send_codecs_to(self, participant):
+        self.send_codecs_to.remote(participant)
+
+
+    def send_codecs_to(self, participant):
+        codecs = self.fsstream.get_property("negotiated-codecs")
+        print "sending stream %s codecs from %s to %s" % \
+              (self.id, self.participant.id, participant.id)
+        if codecs:
+            participant.connect.send_codecs(participant.id, self.id, codecs,
+                                            self.participant.id)            
 
 
 class FsUIParticipant:
@@ -454,12 +595,9 @@ class FsUIParticipant:
     def candidates_done(self, media):
         "Callback for the network object."
         self.streams[media].candidates_done()
-    def codec(self, media, codec):
+    def codecs(self, media, codecs):
         "Callback for the network object."
-        self.streams[media].codec(codec)
-    def codecs_done(self, media):
-        "Callback for the network object."
-        self.streams[media].codecs_done()
+        self.streams[media].codecs(codecs)
     def send_local_codecs(self):
         "Callback for the network object."
         for id in self.streams:
@@ -472,7 +610,10 @@ class FsUIParticipant:
         self.userframe = self.glade.get_widget("user_frame")
         self.glade.get_widget("frame_label").set_text(self.cname)
         self.glade.signal_autoconnect(self)
-        self.mainui.hbox_add(self.userframe)
+        self.label = gtk.Label()
+        self.label.set_alignment(0,0)
+        self.label.show()
+        self.mainui.hbox_add(self.userframe, self.label)
         gtk.gdk.threads_leave()
 
     def exposed(self, widget, *args):
@@ -483,7 +624,7 @@ class FsUIParticipant:
         if not VIDEO:
             return
         try:
-            self.videosink.get_by_name("uservideosink").expose()
+            self.videosink.get_by_interface(gst.interfaces.XOverlay).expose()
         except AttributeError:
             try:
                 self.outcv.acquire()
@@ -539,8 +680,10 @@ class FsUIParticipant:
             except AttributeError:
                 pass
             self.glade.get_widget("user_drawingarea").disconnect_by_func(self.exposed)
-            del self.streams
+            self.streams = {}
             self.outcv.acquire()
+            self.videosink.set_locked_state(True)
+            self.funnel.set_locked_state(True)
             self.videosink.set_state(gst.STATE_NULL)
             self.funnel.set_state(gst.STATE_NULL)
             self.pipeline.pipeline.remove(self.videosink)
@@ -550,6 +693,7 @@ class FsUIParticipant:
             self.outcv.release()
         gtk.gdk.threads_enter()
         self.userframe.destroy()
+        self.label.destroy()
         gtk.gdk.threads_leave()
 
     def error(self):
@@ -559,6 +703,28 @@ class FsUIParticipant:
         else:
             print "ERROR ON %d" % (self.id)
 
+    def recv_codecs_changed(self):
+        codecs = {}
+        for s in self.streams:
+            codec = self.streams[s].fsstream.get_property("current-recv-codecs")
+            mediatype = self.streams[s].session.fssession.get_property("media-type")
+            if len(codec):
+                if mediatype in codecs:
+                    codecs[mediatype] += codec
+                else:
+                    codecs[mediatype] = codec
+        str = ""
+        for mt in codecs:
+            str += "<big>" +mt.value_nick.title() + "</big>:\n"
+            for c in codecs[mt]:
+                str += "  <b>%s</b>: %s %s\n" % (c.id, 
+                                                 c.encoding_name,
+                                                 c.clock_rate)
+        self.label.set_markup(str)
+
+    def send_codecs_to(self, participant):
+        for sid in self.streams:
+            self.streams[sid].send_codecs_to(participant)
     
 
 class FsMainUI:
@@ -567,9 +733,26 @@ class FsMainUI:
     def __init__(self, mode, ip, port):
         self.mode = mode
         self.pipeline = FsUIPipeline()
+        self.pipeline.codecs_changed_audio = self.reset_audio_codecs
+        self.pipeline.codecs_changed_video = self.reset_video_codecs
         self.glade = gtk.glade.XML(gladefile, "main_window")
         self.glade.signal_autoconnect(self)
         self.mainwindow = self.glade.get_widget("main_window")
+        self.audio_combobox = self.glade.get_widget("audio_combobox")
+        self.video_combobox = self.glade.get_widget("video_combobox")
+        liststore = gtk.ListStore(gobject.TYPE_STRING, gobject.TYPE_PYOBJECT)
+        self.audio_combobox.set_model(liststore)
+        cell = gtk.CellRendererText()
+        self.audio_combobox.pack_start(cell, True)
+        self.audio_combobox.add_attribute(cell, 'text', 0)
+        self.reset_audio_codecs()
+        liststore = gtk.ListStore(gobject.TYPE_STRING, gobject.TYPE_PYOBJECT)
+        self.video_combobox.set_model(liststore)
+        cell = gtk.CellRendererText()
+        self.video_combobox.pack_start(cell, True)
+        self.video_combobox.add_attribute(cell, 'text', 0)
+        self.reset_video_codecs()
+
         if mode == CLIENT:
             self.client = FsUIClient(ip, port, mycname, FsUIParticipant,
                                      self.pipeline, self)
@@ -585,12 +768,50 @@ class FsMainUI:
         
         self.mainwindow.show()
 
+    def reset_codecs(self, combobox, fssession):
+        liststore = combobox.get_model()
+        current = fssession.get_property("current-send-codec")
+        liststore.clear()
+        for c in fssession.get_property("codecs"):
+            str = ("%s: %s/%s %s" % (c.id, 
+                                     c.media_type.value_nick,
+                                     c.encoding_name,
+                                     c.clock_rate))
+            iter = liststore.append([str, c])
+            if current and c and current.id == c.id:
+                combobox.set_active_iter(iter)
+                print "active: "+ c.to_string()
+
+    def reset_audio_codecs(self):
+        if AUDIO:
+            self.reset_codecs(self.audio_combobox,
+                              self.pipeline.audiosession.fssession)
+
+    def reset_video_codecs(self):
+        if VIDEO:
+            self.reset_codecs(self.video_combobox,
+                              self.pipeline.videosession.fssession)
+
+    def combobox_changed_cb(self, combobox, fssession):
+        liststore = combobox.get_model()
+        iter = combobox.get_active_iter()
+        if iter:
+            codec = liststore.get_value(iter, 1)
+            fssession.set_send_codec(codec)
+
+    def audio_combobox_changed_cb(self, combobox):
+        self.combobox_changed_cb(combobox, self.pipeline.audiosession.fssession)
+    
+    def video_combobox_changed_cb(self, combobox):
+        self.combobox_changed_cb(combobox, self.pipeline.videosession.fssession)
+        
+        
     def exposed(self, widget, *args):
         "Callback from the exposed event of the widget to make the preview sink"
         if not VIDEO:
             return
         try:
-            self.preview.get_by_name("previewvideosink").expose()
+            self.preview.get_by_interface(gst.interfaces.XOverlay).expose()
         except AttributeError:
             self.preview = self.pipeline.make_video_preview(widget.window.xid,
                                                             self.newsize)
@@ -601,8 +822,11 @@ class FsMainUI:
     def shutdown(self, widget=None):
         gtk.main_quit()
         
-    def hbox_add(self, widget):
-        self.glade.get_widget("user_hbox").pack_start(widget, True, True, 0)
+    def hbox_add(self, widget, label):
+        table = self.glade.get_widget("users_table")
+        x = table.get_properties("n-columns")[0]
+        table.attach(widget, x, x+1, 0, 1)
+        table.attach(label, x, x+1, 1, 3, xpadding=6)
 
     def __del__(self):
         self.mainwindow.destroy()
@@ -619,6 +843,35 @@ class FsMainUI:
         gtk.main_quit()
         gtk.gdk.threads_leave()
 
+    def show_dtmf(self, button):
+        try:
+            self.dtmf.present()
+        except AttributeError:
+            self.dtmf =  gtk.glade.XML(gladefile, "dtmf_window")
+            self.dtmf.signal_autoconnect(self)
+
+    def dtmf_start(self, button):
+        if (self.dtmf.get_widget("dtmf_as_event").get_active()):
+            self.dtmf_last_method = farsight.DTMF_METHOD_RTP_RFC4733
+        elif (self.dtmf.get_widget("dtmf_as_sound").get_active()):
+            self.dtmf_last_method = farsight.DTMF_METHOD_IN_BAND
+        else:
+            print "Invalid DTMF Method"
+            return
+        self.pipeline.audiosession.dtmf_start(button.get_label(), \
+                                              self.dtmf_last_method)
+                                              
+    def dtmf_stop(self, button):
+        try:
+            self.pipeline.audiosession.dtmf_stop(self.dtmf_last_method)
+            del self.dtmf_last_method
+        except AttributeError:
+            pass
+    def dtmf_destroy(self, button):
+        self.dtmf.get_widget("dtmf_window").destroy()
+        del self.dtmf
+
+
 
 class FsUIStartup:
     "Displays the startup window and then creates the FsMainUI"
