@@ -5,7 +5,7 @@
  *  @author: Olivier Crete <olivier.crete@collabora.co.uk>
  * Copyright 2007 Nokia Corp.
  *
- * fs-multicast-transmitter.c - A Farsight Multiast UDP stream transmitter
+ * fs-multicast-stream-transmitter.c - A Farsight Multiast UDP stream transmitter
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -26,41 +26,31 @@
 /**
  * SECTION:fs-multicast-stream-transmitter
  * @short_description: A stream transmitter object for Multicast UDP
- * @see_also: fs-rawudp-stream-transmitter
+ * @see_also: #FsRawUdpStreamTransmitter
  *
- * <refsect2>
- * <para>
  * The multicast transmitter allows data to be sent over and received from
  * multicasted UDP on IPv4.
- * </para>
  *
- * <para>
  * This stream transmitter never emits local candidates. It will listen
  * to the port specified in the remote candidate. And will also send to that
  * port. It accepts only a single remote candidate per component, if a new one
  * is given, it will replace the previous one for that component.
- * </para>
  *
- * <para>
  * The transmitter will only stop sending to a multicast group when all of its
  * StreamTransmitters that have this multicast group as destination have their
  * "sending" property set to false. Multiple stream transmitters can point to
  * the same multicast groups from the same Transmitter (session), and only one
  * copy of each packet will be received.
- * </para>
  *
- * <para>
  * It will only listen to and send from the IP specified in the
  * prefered-local-candidates. There can be only one preferred candidate per
  * component. Only the component_id and the ip will be used from the preferred
  * local candidates, everything else is ignored.
- * </para>
  *
- * <para>
  * Packets sent will be looped back (so that other clients on the same session
  * can be on the same machine.
- * </para>
- * </refsect2>
+ *
+ * The name of this transmitter is "multicast".
  */
 
 #ifdef HAVE_CONFIG_H
@@ -77,11 +67,19 @@
 
 #include <string.h>
 #include <sys/types.h>
-#include <sys/socket.h>
-#include <netinet/in.h>
-#include <arpa/inet.h>
-#include <netdb.h>
-#include <unistd.h>
+
+#ifdef HAVE_UNISTD_H
+# include <unistd.h>
+#endif
+
+#ifdef G_OS_WIN32
+# include <winsock2.h>
+#else /*G_OS_WIN32*/
+# include <netdb.h>
+# include <sys/socket.h>
+# include <netinet/in.h>
+# include <arpa/inet.h>
+#endif /*G_OS_WIN32*/
 
 GST_DEBUG_CATEGORY_EXTERN (fs_multicast_transmitter_debug);
 #define GST_CAT_DEFAULT fs_multicast_transmitter_debug
@@ -110,26 +108,33 @@ struct _FsMulticastStreamTransmitterPrivate
    */
   FsMulticastTransmitter *transmitter;
 
+  GMutex *mutex;
+
+  /* Protected by the mutex */
   gboolean sending;
 
   /*
    * We have at most of those per component (index 0 is unused)
    */
-  FsCandidate **remote_candidate;
   FsCandidate **local_candidate;
+  /* Protected by the mutex */
+  FsCandidate **remote_candidate;
 
-  UdpSock **recvudpsocks;
-  UdpSock **sendudpsocks;
+  /* Protected by the mutex */
+  UdpSock **udpsocks;
 
   GList *preferred_local_candidates;
-
-  guint next_candidate_id;
 };
 
 #define FS_MULTICAST_STREAM_TRANSMITTER_GET_PRIVATE(o)  \
   (G_TYPE_INSTANCE_GET_PRIVATE ((o), FS_TYPE_MULTICAST_STREAM_TRANSMITTER, \
                                 FsMulticastStreamTransmitterPrivate))
 
+#define FS_MULTICAST_STREAM_TRANSMITTER_LOCK(s) \
+  g_mutex_lock ((s)->priv->mutex)
+#define FS_MULTICAST_STREAM_TRANSMITTER_UNLOCK(s) \
+  g_mutex_unlock ((s)->priv->mutex)
+
 static void fs_multicast_stream_transmitter_class_init (FsMulticastStreamTransmitterClass *klass);
 static void fs_multicast_stream_transmitter_init (FsMulticastStreamTransmitter *self);
 static void fs_multicast_stream_transmitter_dispose (GObject *object);
@@ -144,8 +149,8 @@ static void fs_multicast_stream_transmitter_set_property (GObject *object,
                                                 const GValue *value,
                                                 GParamSpec *pspec);
 
-static gboolean fs_multicast_stream_transmitter_add_remote_candidate (
-    FsStreamTransmitter *streamtransmitter, FsCandidate *candidate,
+static gboolean fs_multicast_stream_transmitter_set_remote_candidates (
+    FsStreamTransmitter *streamtransmitter, GList *candidates,
     GError **error);
 
 
@@ -193,8 +198,8 @@ fs_multicast_stream_transmitter_class_init (FsMulticastStreamTransmitterClass *k
   gobject_class->set_property = fs_multicast_stream_transmitter_set_property;
   gobject_class->get_property = fs_multicast_stream_transmitter_get_property;
 
-  streamtransmitterclass->add_remote_candidate =
-    fs_multicast_stream_transmitter_add_remote_candidate;
+  streamtransmitterclass->set_remote_candidates =
+    fs_multicast_stream_transmitter_set_remote_candidates;
 
   g_object_class_override_property (gobject_class, PROP_SENDING, "sending");
   g_object_class_override_property (gobject_class,
@@ -214,6 +219,8 @@ fs_multicast_stream_transmitter_init (FsMulticastStreamTransmitter *self)
   self->priv->disposed = FALSE;
 
   self->priv->sending = TRUE;
+
+  self->priv->mutex = g_mutex_new ();
 }
 
 static void
@@ -226,31 +233,18 @@ fs_multicast_stream_transmitter_dispose (GObject *object)
     /* If dispose did already run, return. */
     return;
 
-  if (self->priv->sendudpsocks)
+  if (self->priv->udpsocks)
   {
     for (c = 1; c <= self->priv->transmitter->components; c++)
     {
-      if (self->priv->sendudpsocks[c])
+      if (self->priv->udpsocks[c])
       {
         if (self->priv->sending)
           fs_multicast_transmitter_udpsock_dec_sending (
-              self->priv->sendudpsocks[c]);
+              self->priv->udpsocks[c]);
         fs_multicast_transmitter_put_udpsock (self->priv->transmitter,
-            self->priv->sendudpsocks[c]);
-        self->priv->sendudpsocks[c] = NULL;
-      }
-    }
-  }
-
-  if (self->priv->recvudpsocks)
-  {
-    for (c = 1; c <= self->priv->transmitter->components; c++)
-    {
-      if (self->priv->recvudpsocks[c])
-      {
-        fs_multicast_transmitter_put_udpsock (self->priv->transmitter,
-            self->priv->recvudpsocks[c]);
-        self->priv->recvudpsocks[c] = NULL;
+            self->priv->udpsocks[c], self->priv->remote_candidate[c]->ttl);
+        self->priv->udpsocks[c] = NULL;
       }
     }
   }
@@ -297,11 +291,8 @@ fs_multicast_stream_transmitter_finalize (GObject *object)
     self->priv->local_candidate = NULL;
   }
 
-  g_free (self->priv->sendudpsocks);
-  self->priv->sendudpsocks = NULL;
-
-  g_free (self->priv->recvudpsocks);
-  self->priv->recvudpsocks = NULL;
+  g_free (self->priv->udpsocks);
+  self->priv->udpsocks = NULL;
 
   parent_class->finalize (object);
 }
@@ -317,7 +308,9 @@ fs_multicast_stream_transmitter_get_property (GObject *object,
   switch (prop_id)
   {
     case PROP_SENDING:
+      FS_MULTICAST_STREAM_TRANSMITTER_LOCK (self);
       g_value_set_boolean (value, self->priv->sending);
+      FS_MULTICAST_STREAM_TRANSMITTER_UNLOCK (self);
       break;
     case PROP_PREFERRED_LOCAL_CANDIDATES:
       g_value_set_boxed (value, self->priv->preferred_local_candidates);
@@ -340,21 +333,32 @@ fs_multicast_stream_transmitter_set_property (GObject *object,
     case PROP_SENDING:
       {
         gboolean old_sending = self->priv->sending;
+        gboolean sending = g_value_get_boolean (value);
         gint c;
 
-        self->priv->sending = g_value_get_boolean (value);
+        FS_MULTICAST_STREAM_TRANSMITTER_LOCK (self);
+        self->priv->sending = sending;
 
-        if (self->priv->sending != old_sending)
+        if (sending != old_sending)
           for (c = 1; c <= self->priv->transmitter->components; c++)
-            if (self->priv->sendudpsocks[c])
+            if (self->priv->udpsocks[c])
             {
-              if (self->priv->sending)
+              guint8 ttl = self->priv->remote_candidate[c]->ttl;
+              fs_multicast_transmitter_udpsock_ref (self->priv->transmitter,
+                  self->priv->udpsocks[c], ttl);
+              FS_MULTICAST_STREAM_TRANSMITTER_UNLOCK (self);
+              if (sending)
                 fs_multicast_transmitter_udpsock_inc_sending (
-                    self->priv->sendudpsocks[c]);
+                    self->priv->udpsocks[c]);
               else
                 fs_multicast_transmitter_udpsock_dec_sending (
-                    self->priv->sendudpsocks[c]);
+                    self->priv->udpsocks[c]);
+              fs_multicast_transmitter_put_udpsock (self->priv->transmitter,
+                  self->priv->udpsocks[c], ttl);
+              FS_MULTICAST_STREAM_TRANSMITTER_LOCK (self);
             }
+        FS_MULTICAST_STREAM_TRANSMITTER_UNLOCK (self);
+
       }
       break;
     case PROP_PREFERRED_LOCAL_CANDIDATES:
@@ -373,9 +377,7 @@ fs_multicast_stream_transmitter_build (FsMulticastStreamTransmitter *self,
   GList *item;
   gint c;
 
-  self->priv->sendudpsocks = g_new0 (UdpSock *,
-      self->priv->transmitter->components + 1);
-  self->priv->recvudpsocks = g_new0 (UdpSock *,
+  self->priv->udpsocks = g_new0 (UdpSock *,
       self->priv->transmitter->components + 1);
   self->priv->local_candidate = g_new0 (FsCandidate *,
       self->priv->transmitter->components + 1);
@@ -447,64 +449,15 @@ fs_multicast_stream_transmitter_build (FsMulticastStreamTransmitter *self,
   return TRUE;
 }
 
-/**
- * fs_multicast_stream_transmitter_add_remote_candidate
- * @streamtransmitter: a #FsStreamTransmitter
- * @candidate: a remote #FsCandidate to add
- * @error: location of a #GError, or NULL if no error occured
- *
- * This function is used to add remote candidates to the transmitter
- *
- * Returns: TRUE of the candidate could be added, FALSE if it couldnt
- *   (and the #GError will be set)
- */
 
 static gboolean
 fs_multicast_stream_transmitter_add_remote_candidate (
-    FsStreamTransmitter *streamtransmitter, FsCandidate *candidate,
+    FsMulticastStreamTransmitter *self, FsCandidate *candidate,
     GError **error)
 {
-  FsMulticastStreamTransmitter *self =
-    FS_MULTICAST_STREAM_TRANSMITTER (streamtransmitter);
-  UdpSock *newrecvudpsock = NULL;
-  UdpSock *newsendudpsock = NULL;
-
-  if (candidate->proto != FS_NETWORK_PROTOCOL_UDP) {
-    g_set_error (error, FS_ERROR, FS_ERROR_INVALID_ARGUMENTS,
-      "You set a candidate of a type %d that is not  FS_NETWORK_PROTOCOL_UDP",
-      candidate->proto);
-    return FALSE;
-  }
-
-  if (candidate->type != FS_CANDIDATE_TYPE_MULTICAST)
-  {
-    g_set_error (error, FS_ERROR, FS_ERROR_INVALID_ARGUMENTS,
-        "The remote candidate is not of the right type, it should be"
-        " FS_ERROR_INVALID_ARGUMENTS, but it is %d", candidate->type);
-    return FALSE;
-  }
-
-  if (!candidate->ip || !candidate->port) {
-    g_set_error (error, FS_ERROR, FS_ERROR_INVALID_ARGUMENTS,
-      "The candidate passed does not contain a valid ip or port");
-    return FALSE;
-  }
-
-  if (candidate->component_id == 0 ||
-    candidate->component_id > self->priv->transmitter->components) {
-    g_set_error (error, FS_ERROR, FS_ERROR_INVALID_ARGUMENTS,
-      "The candidate passed has has an invalid component id %u (not in [0,%u])",
-      candidate->component_id, self->priv->transmitter->components);
-    return FALSE;
-  }
-
-  if (candidate->ttl == 0)
-  {
-    g_set_error (error, FS_ERROR, FS_ERROR_INVALID_ARGUMENTS,
-        "The TTL for IPv4 multicast candidates must not be 0");
-    return FALSE;
-  }
+  UdpSock *newudpsock = NULL;
 
+  FS_MULTICAST_STREAM_TRANSMITTER_LOCK (self);
   if (self->priv->remote_candidate[candidate->component_id])
   {
     FsCandidate *old_candidate =
@@ -514,71 +467,54 @@ fs_multicast_stream_transmitter_add_remote_candidate (
         !strcmp (old_candidate->ip, candidate->ip))
     {
       GST_DEBUG ("Re-set the same candidate, ignoring");
-      return TRUE;
+      FS_MULTICAST_STREAM_TRANSMITTER_UNLOCK (self);
+     return TRUE;
     }
-    fs_candidate_destroy (old_candidate);
-    self->priv->remote_candidate[candidate->component_id] = NULL;
   }
 
+  FS_MULTICAST_STREAM_TRANSMITTER_UNLOCK (self);
+
   /*
    * IMPROVE ME: We should probably check that the candidate's IP
    *  has the format x.x.x.x where x is [0,255] using GRegex, etc
    * We should also check if the address is in the multicast range
    */
 
-  newrecvudpsock = fs_multicast_transmitter_get_udpsock (
+  newudpsock = fs_multicast_transmitter_get_udpsock (
       self->priv->transmitter,
       candidate->component_id,
       self->priv->local_candidate[candidate->component_id]->ip,
       candidate->ip,
       candidate->port,
       candidate->ttl,
-      TRUE, error);
+      error);
 
-  if (!newrecvudpsock)
+  if (!newudpsock)
     return FALSE;
 
-  newsendudpsock = fs_multicast_transmitter_get_udpsock (
-      self->priv->transmitter,
-      candidate->component_id,
-      self->priv->local_candidate[candidate->component_id]->ip,
-      candidate->ip,
-      candidate->port,
-      candidate->ttl,
-      FALSE, error);
+  FS_MULTICAST_STREAM_TRANSMITTER_LOCK (self);
 
-  if (!newsendudpsock)
-  {
-    fs_multicast_transmitter_put_udpsock (self->priv->transmitter,
-        newrecvudpsock);
-    return FALSE;
-  }
-
-  if (self->priv->sendudpsocks[candidate->component_id])
+  if (self->priv->udpsocks[candidate->component_id])
   {
     if (self->priv->sending)
       fs_multicast_transmitter_udpsock_dec_sending (
-          self->priv->sendudpsocks[candidate->component_id]);
+          self->priv->udpsocks[candidate->component_id]);
     fs_multicast_transmitter_put_udpsock (self->priv->transmitter,
-        self->priv->sendudpsocks[candidate->component_id]);
+        self->priv->udpsocks[candidate->component_id],
+        self->priv->remote_candidate[candidate->component_id]->ttl);
   }
 
-  if (self->priv->recvudpsocks[candidate->component_id])
-    fs_multicast_transmitter_put_udpsock (self->priv->transmitter,
-        self->priv->recvudpsocks[candidate->component_id]);
-
-  self->priv->sendudpsocks[candidate->component_id] = newsendudpsock;
-  self->priv->recvudpsocks[candidate->component_id] = newrecvudpsock;
+  self->priv->udpsocks[candidate->component_id] = newudpsock;
 
   if (self->priv->sending)
     fs_multicast_transmitter_udpsock_inc_sending (
-        self->priv->sendudpsocks[candidate->component_id]);
+        self->priv->udpsocks[candidate->component_id]);
 
+  fs_candidate_destroy (self->priv->remote_candidate[candidate->component_id]);
   self->priv->remote_candidate[candidate->component_id] =
     fs_candidate_copy (candidate);
 
-  self->priv->local_candidate[candidate->component_id]->port = candidate->port;
-
+  FS_MULTICAST_STREAM_TRANSMITTER_UNLOCK (self);
 
   g_signal_emit_by_name (self, "new-active-candidate-pair",
       self->priv->local_candidate[candidate->component_id],
@@ -587,6 +523,70 @@ fs_multicast_stream_transmitter_add_remote_candidate (
   return TRUE;
 }
 
+/**
+ * fs_multicast_stream_transmitter_set_remote_candidates
+ */
+
+static gboolean
+fs_multicast_stream_transmitter_set_remote_candidates (
+    FsStreamTransmitter *streamtransmitter, GList *candidates,
+    GError **error)
+{
+  GList *item = NULL;
+  FsMulticastStreamTransmitter *self =
+    FS_MULTICAST_STREAM_TRANSMITTER (streamtransmitter);
+
+  for (item = candidates; item; item = g_list_next (item))
+  {
+    FsCandidate *candidate = item->data;
+
+    if (candidate->proto != FS_NETWORK_PROTOCOL_UDP) {
+      g_set_error (error, FS_ERROR, FS_ERROR_INVALID_ARGUMENTS,
+          "You set a candidate of a type %d that is not"
+          " FS_NETWORK_PROTOCOL_UDP",
+          candidate->proto);
+      return FALSE;
+    }
+
+    if (candidate->type != FS_CANDIDATE_TYPE_MULTICAST)
+    {
+      g_set_error (error, FS_ERROR, FS_ERROR_INVALID_ARGUMENTS,
+          "The remote candidate is not of the right type, it should be"
+          " FS_ERROR_INVALID_ARGUMENTS, but it is %d", candidate->type);
+      return FALSE;
+    }
+
+    if (!candidate->ip || !candidate->port) {
+      g_set_error (error, FS_ERROR, FS_ERROR_INVALID_ARGUMENTS,
+          "The candidate passed does not contain a valid ip or port");
+      return FALSE;
+    }
+
+    if (candidate->component_id == 0 ||
+        candidate->component_id > self->priv->transmitter->components) {
+      g_set_error (error, FS_ERROR, FS_ERROR_INVALID_ARGUMENTS,
+          "The candidate passed has an invalid component id %u (not in [1,%u])",
+          candidate->component_id, self->priv->transmitter->components);
+      return FALSE;
+    }
+
+    if (candidate->ttl == 0)
+    {
+      g_set_error (error, FS_ERROR, FS_ERROR_INVALID_ARGUMENTS,
+          "The TTL for IPv4 multicast candidates must not be 0");
+      return FALSE;
+    }
+  }
+
+  for (item = candidates; item; item = g_list_next (item))
+    if (!fs_multicast_stream_transmitter_add_remote_candidate (self,
+            item->data, error))
+      return FALSE;
+
+
+  return TRUE;
+}
+
 
 FsMulticastStreamTransmitter *
 fs_multicast_stream_transmitter_newv (FsMulticastTransmitter *transmitter,
