@@ -5,7 +5,7 @@
  *  @author: Olivier Crete <olivier.crete@collabora.co.uk>
  * Copyright 2007-2008 Nokia Corp.
  *
- * fs-multicast-transmitter.h - A Farsight multicast UDP transmitter
+ * fs-multicast-transmitter.c - A Farsight multicast UDP transmitter
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -42,11 +42,20 @@
 
 #include <string.h>
 #include <sys/types.h>
-#include <sys/socket.h>
-#include <netinet/in.h>
-#include <arpa/inet.h>
-#include <netdb.h>
-#include <unistd.h>
+
+#ifdef HAVE_UNISTD_H
+# include <unistd.h>
+#endif
+
+#ifdef G_OS_WIN32
+# include <ws2tcpip.h>
+# define close closesocket
+#else /*G_OS_WIN32*/
+# include <netdb.h>
+# include <sys/socket.h>
+# include <netinet/in.h>
+# include <arpa/inet.h>
+#endif /*G_OS_WIN32*/
 
 GST_DEBUG_CATEGORY (fs_multicast_transmitter_debug);
 #define GST_CAT_DEFAULT fs_multicast_transmitter_debug
@@ -78,6 +87,7 @@ struct _FsMulticastTransmitterPrivate
   GstElement **udpsrc_funnels;
   GstElement **udpsink_tees;
 
+  GMutex *mutex;
   GList **udpsocks;
 
   gboolean disposed;
@@ -107,8 +117,7 @@ static FsStreamTransmitter *fs_multicast_transmitter_new_stream_transmitter (
     FsTransmitter *transmitter, FsParticipant *participant,
     guint n_parameters, GParameter *parameters, GError **error);
 static GType fs_multicast_transmitter_get_stream_transmitter_type (
-    FsTransmitter *transmitter,
-    GError **error);
+    FsTransmitter *transmitter);
 
 static GObjectClass *parent_class = NULL;
 //static guint signals[LAST_SIGNAL] = { 0 };
@@ -142,10 +151,9 @@ fs_multicast_transmitter_register_type (FsPlugin *module)
     (GInstanceInitFunc) fs_multicast_transmitter_init
   };
 
-  if (fs_multicast_transmitter_debug == NULL)
-    GST_DEBUG_CATEGORY_INIT (fs_multicast_transmitter_debug,
-        "fsmulticasttransmitter", 0,
-        "Farsight multicast UDP transmitter");
+  GST_DEBUG_CATEGORY_INIT (fs_multicast_transmitter_debug,
+      "fsmulticasttransmitter", 0,
+      "Farsight multicast UDP transmitter");
 
   fs_multicast_stream_transmitter_register_type (module);
 
@@ -155,18 +163,7 @@ fs_multicast_transmitter_register_type (FsPlugin *module)
   return type;
 }
 
-static void
-fs_multicast_transmitter_unload (FsPlugin *plugin)
-{
-  if (fs_multicast_transmitter_debug)
-  {
-    gst_debug_category_free (fs_multicast_transmitter_debug);
-    fs_multicast_transmitter_debug = NULL;
-  }
-}
-
-FS_INIT_PLUGIN (fs_multicast_transmitter_register_type,
-    fs_multicast_transmitter_unload)
+FS_INIT_PLUGIN (fs_multicast_transmitter_register_type)
 
 static void
 fs_multicast_transmitter_class_init (FsMulticastTransmitterClass *klass)
@@ -206,6 +203,7 @@ fs_multicast_transmitter_init (FsMulticastTransmitter *self)
   self->priv->disposed = FALSE;
 
   self->components = 2;
+  self->priv->mutex = g_mutex_new ();
 }
 
 static void
@@ -352,6 +350,8 @@ fs_multicast_transmitter_constructed (GObject *object)
       return;
     }
   }
+
+  GST_CALL_PARENT (G_OBJECT_CLASS, constructed, (object));
 }
 
 static void
@@ -400,6 +400,8 @@ fs_multicast_transmitter_finalize (GObject *object)
     self->priv->udpsocks = NULL;
   }
 
+  g_mutex_free (self->priv->mutex);
+
   parent_class->finalize (object);
 }
 
@@ -418,6 +420,9 @@ fs_multicast_transmitter_get_property (GObject *object,
     case PROP_GST_SRC:
       g_value_set_object (value, self->priv->gst_src);
       break;
+    case PROP_COMPONENTS:
+      g_value_set_uint (value, self->components);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -469,13 +474,12 @@ fs_multicast_transmitter_new_stream_transmitter (FsTransmitter *transmitter,
 
 /*
  * The UdpSock structure is a ref-counted pseudo-object use to represent
- * one local_ip:port:ttl:multicast_ip quatuor on which we listen and send,
+ * one local_ip:port:multicast_ip trio on which we listen and send,
  * so it includes a udpsrc and a multiudpsink. It represents one BSD socket.
- * If two UdpSock only differ by their TTL, only the first will have 
+ * The TTL used is the max TTL requested by any stream.
  */
 
 struct _UdpSock {
-  gint refcount;
 
   GstElement *udpsrc;
   GstPad *udpsrc_requested_pad;
@@ -486,18 +490,21 @@ struct _UdpSock {
   gchar *local_ip;
   gchar *multicast_ip;
   guint16 port;
-  guint8 ttl;
+  /* Protected by the transmitter mutex */
+  guint8 current_ttl;
 
   gint fd;
 
+  /* Protected by the transmitter mutex */
+  GByteArray *ttls;
+
   /* These are just convenience pointers to our parent transmitter */
   GstElement *funnel;
   GstElement *tee;
 
   guint component_id;
 
-  gboolean sendonly;
-  gint sendcount;
+  volatile gint sendcount;
 };
 
 static gboolean
@@ -518,7 +525,7 @@ _ip_string_into_sockaddr_in (const gchar *ip_as_string,
         gai_strerror (retval));
     return FALSE;
   }
-  memcpy (sockaddr_in, result->ai_addr, sizeof(struct sockaddr_in));
+  memcpy (sockaddr_in, result->ai_addr, sizeof (struct sockaddr_in));
   freeaddrinfo (result);
 
   return TRUE;
@@ -537,7 +544,11 @@ _bind_port (
   int retval;
   guchar loop = 1;
   int reuseaddr = 1;
-  struct ip_mreqn mreqn;
+#ifdef HAVE_IP_MREQN
+  struct ip_mreqn mreq;
+#else
+  struct ip_mreq mreq;
+#endif
 
   address.sin_family = AF_INET;
   address.sin_addr.s_addr = INADDR_ANY;
@@ -546,20 +557,32 @@ _bind_port (
 
   if (!_ip_string_into_sockaddr_in (multicast_ip, &address, error))
     goto error;
-  memcpy (&mreqn.imr_multiaddr, &address.sin_addr, sizeof(mreqn.imr_multiaddr));
+  memcpy (&mreq.imr_multiaddr, &address.sin_addr,
+      sizeof (mreq.imr_multiaddr));
 
   if (local_ip)
   {
     struct sockaddr_in tmpaddr;
     if (!_ip_string_into_sockaddr_in (local_ip, &tmpaddr, error))
       goto error;
-    memcpy (&mreqn.imr_address, &tmpaddr.sin_addr, sizeof(mreqn.imr_address));
+#ifdef HAVE_IP_MREQN
+    memcpy (&mreq.imr_address, &tmpaddr.sin_addr, sizeof (mreq.imr_address));
+#else
+    memcpy (&mreq.imr_interface, &tmpaddr.sin_addr, sizeof (mreq.imr_interface));
+#endif
   }
   else
   {
-    mreqn.imr_address.s_addr = INADDR_ANY;
+#ifdef HAVE_IP_MREQN
+    mreq.imr_address.s_addr = INADDR_ANY;
+#else
+    mreq.imr_interface.s_addr = INADDR_ANY;
+#endif
   }
-  mreqn.imr_ifindex = 0;
+
+#ifdef HAVE_IP_MREQN
+  mreq.imr_ifindex = 0;
+#endif
 
   if ((sock = socket (AF_INET, SOCK_DGRAM, IPPROTO_UDP)) <= 0) {
     g_set_error (error, FS_ERROR, FS_ERROR_NETWORK,
@@ -567,7 +590,7 @@ _bind_port (
     goto error;
   }
 
-  if (setsockopt (sock, IPPROTO_IP, IP_MULTICAST_TTL, &ttl,
+  if (setsockopt (sock, IPPROTO_IP, IP_MULTICAST_TTL, (const void *)&ttl,
           sizeof (ttl)) < 0)
   {
     g_set_error (error, FS_ERROR, FS_ERROR_INVALID_ARGUMENTS,
@@ -576,7 +599,7 @@ _bind_port (
     goto error;
   }
 
-  if (setsockopt (sock, IPPROTO_IP, IP_MULTICAST_LOOP, &loop,
+  if (setsockopt (sock, IPPROTO_IP, IP_MULTICAST_LOOP, (const void *)&loop,
           sizeof (loop)) < 0)
   {
     g_set_error (error, FS_ERROR, FS_ERROR_INVALID_ARGUMENTS,
@@ -585,7 +608,7 @@ _bind_port (
     goto error;
   }
 
-  if (setsockopt (sock, SOL_SOCKET, SO_REUSEADDR, &reuseaddr,
+  if (setsockopt (sock, SOL_SOCKET, SO_REUSEADDR, (const void *)&reuseaddr,
           sizeof (reuseaddr)) < 0)
   {
     g_set_error (error, FS_ERROR, FS_ERROR_INVALID_ARGUMENTS,
@@ -595,7 +618,7 @@ _bind_port (
   }
 
 #ifdef SO_REUSEPORT
-  if (setsockopt (sock, SOL_SOCKET, SO_REUSEPORT, &reuseaddr,
+  if (setsockopt (sock, SOL_SOCKET, SO_REUSEPORT, (const void *)&reuseaddr,
           sizeof (reuseaddr)) < 0)
   {
     g_set_error (error, FS_ERROR, FS_ERROR_INVALID_ARGUMENTS,
@@ -606,7 +629,7 @@ _bind_port (
 #endif
 
   if (setsockopt (sock, IPPROTO_IP, IP_ADD_MEMBERSHIP,
-          &(mreqn), sizeof (mreqn)) < 0)
+          (const void *)&mreq, sizeof (mreq)) < 0)
   {
     g_set_error (error, FS_ERROR, FS_ERROR_INVALID_ARGUMENTS,
         "Could not join the socket to the multicast group: %s",
@@ -655,7 +678,8 @@ _create_sinksource (gchar *elementname, GstBin *bin,
     "sockfd", fd,
     NULL);
 
-  if (direction == GST_PAD_SINK)
+  if (g_object_class_find_property (G_OBJECT_GET_CLASS (elem),
+          "auto-multicast"))
     g_object_set (elem, "auto-multicast", FALSE, NULL);
 
   if (!gst_bin_add (bin, elem)) {
@@ -722,27 +746,17 @@ _create_sinksource (gchar *elementname, GstBin *bin,
   return NULL;
 }
 
-UdpSock *
-fs_multicast_transmitter_get_udpsock (FsMulticastTransmitter *trans,
+static UdpSock *
+fs_multicast_transmitter_get_udpsock_locked (FsMulticastTransmitter *trans,
     guint component_id,
     const gchar *local_ip,
     const gchar *multicast_ip,
     guint16 port,
     guint8 ttl,
-    gboolean recv,
     GError **error)
 {
   UdpSock *udpsock;
   GList *udpsock_e;
-  gboolean sendonly = FALSE;
-
-  /* First lets check if we already have one */
-  if (component_id > trans->components)
-  {
-    g_set_error (error, FS_ERROR, FS_ERROR_INVALID_ARGUMENTS,
-      "Invalid component %d > %d", component_id, trans->components);
-    return NULL;
-  }
 
   for (udpsock_e = g_list_first (trans->priv->udpsocks[component_id]);
        udpsock_e;
@@ -755,36 +769,65 @@ fs_multicast_transmitter_get_udpsock (FsMulticastTransmitter *trans,
         ((local_ip == NULL && udpsock->local_ip == NULL) ||
           !strcmp (local_ip, udpsock->local_ip)))
     {
-      if (recv)
+      if (ttl > udpsock->current_ttl)
       {
-        if (!udpsock->sendonly)
+
+        if (setsockopt (udpsock->fd, IPPROTO_IP, IP_MULTICAST_TTL,
+                (const void *)&ttl, sizeof (ttl)) < 0)
         {
-          udpsock->refcount++;
-          return udpsock;
+          g_set_error (error, FS_ERROR, FS_ERROR_INVALID_ARGUMENTS,
+              "Error setting the multicast TTL: %s",
+              g_strerror (errno));
+          return NULL;
         }
+        udpsock->current_ttl = ttl;
       }
-      else /* recv */
-      {
-        if (ttl == udpsock->ttl)
-        {
-          udpsock->refcount++;
-          return udpsock;
-        }
-        sendonly = TRUE;
-      }
+      g_byte_array_append (udpsock->ttls, &ttl, 1);
+      g_mutex_unlock (trans->priv->mutex);
+      return udpsock;
     }
   }
+  return NULL;
+}
 
-  udpsock = g_new0 (UdpSock, 1);
+UdpSock *
+fs_multicast_transmitter_get_udpsock (FsMulticastTransmitter *trans,
+    guint component_id,
+    const gchar *local_ip,
+    const gchar *multicast_ip,
+    guint16 port,
+    guint8 ttl,
+    GError **error)
+{
+  UdpSock *udpsock;
+  UdpSock *tmpudpsock;
+
+  /* First lets check if we already have one */
+  if (component_id > trans->components)
+  {
+    g_set_error (error, FS_ERROR, FS_ERROR_INVALID_ARGUMENTS,
+      "Invalid component %d > %d", component_id, trans->components);
+    return NULL;
+  }
+
+  g_mutex_lock (trans->priv->mutex);
+  udpsock = fs_multicast_transmitter_get_udpsock_locked (trans, component_id,
+      local_ip, multicast_ip, port, ttl, error);
+  g_mutex_unlock (trans->priv->mutex);
+
+  if (udpsock)
+    return udpsock;
+
+  udpsock = g_slice_new0 (UdpSock);
 
-  udpsock->refcount = 1;
   udpsock->local_ip = g_strdup (local_ip);
   udpsock->multicast_ip = g_strdup (multicast_ip);
   udpsock->fd = -1;
   udpsock->component_id = component_id;
   udpsock->port = port;
-  udpsock->ttl = ttl;
-  udpsock->sendonly = sendonly;
+  udpsock->current_ttl = ttl;
+  udpsock->ttls = g_byte_array_new ();
+  g_byte_array_append (udpsock->ttls, &ttl, 1);
 
   /* Now lets bind both ports */
 
@@ -795,17 +838,13 @@ fs_multicast_transmitter_get_udpsock (FsMulticastTransmitter *trans,
   /* Now lets create the elements */
 
   udpsock->tee = trans->priv->udpsink_tees[component_id];
-  if (!udpsock->sendonly)
-    udpsock->funnel = trans->priv->udpsrc_funnels[component_id];
+  udpsock->funnel = trans->priv->udpsrc_funnels[component_id];
 
-  if (!udpsock->sendonly)
-  {
-    udpsock->udpsrc = _create_sinksource ("udpsrc",
-        GST_BIN (trans->priv->gst_src), udpsock->funnel, udpsock->fd,
-        GST_PAD_SRC, &udpsock->udpsrc_requested_pad, error);
-    if (!udpsock->udpsrc)
-      goto error;
-  }
+  udpsock->udpsrc = _create_sinksource ("udpsrc",
+      GST_BIN (trans->priv->gst_src), udpsock->funnel, udpsock->fd,
+      GST_PAD_SRC, &udpsock->udpsrc_requested_pad, error);
+  if (!udpsock->udpsrc)
+    goto error;
 
   udpsock->udpsink = _create_sinksource ("multiudpsink",
     GST_BIN (trans->priv->gst_sink), udpsock->tee, udpsock->fd, GST_PAD_SINK,
@@ -818,31 +857,85 @@ fs_multicast_transmitter_get_udpsock (FsMulticastTransmitter *trans,
       "sync", FALSE,
       NULL);
 
+  g_mutex_lock (trans->priv->mutex);
+  /* Check if someone else has added the same thing at the same time */
+  tmpudpsock = fs_multicast_transmitter_get_udpsock_locked (trans, component_id,
+      local_ip, multicast_ip, port, ttl, error);
+
+  if (tmpudpsock)
+  {
+    g_mutex_unlock (trans->priv->mutex);
+    fs_multicast_transmitter_put_udpsock (trans, udpsock, ttl);
+    return tmpudpsock;
+  }
+
   trans->priv->udpsocks[component_id] =
     g_list_prepend (trans->priv->udpsocks[component_id], udpsock);
+  g_mutex_unlock (trans->priv->mutex);
 
   return udpsock;
 
  error:
 
   if (udpsock)
-    fs_multicast_transmitter_put_udpsock (trans, udpsock);
+    fs_multicast_transmitter_put_udpsock (trans, udpsock, ttl);
 
   return NULL;
 }
 
 void
 fs_multicast_transmitter_put_udpsock (FsMulticastTransmitter *trans,
-  UdpSock *udpsock)
+    UdpSock *udpsock, guint8 ttl)
 {
-  if (udpsock->refcount > 1) {
-    udpsock->refcount--;
+  guint i;
+
+  g_mutex_lock (trans->priv->mutex);
+  for (i = udpsock->ttls->len - 1;; i--)
+  {
+    if (udpsock->ttls->data[i] == ttl)
+    {
+      g_byte_array_remove_index_fast (udpsock->ttls, i);
+      break;
+    }
+
+    g_return_if_fail (i > 0);
+  }
+
+  if (udpsock->ttls->len > 0)
+  {
+    /* If we were the max, check if there is a new max */
+    if (udpsock->current_ttl == ttl && ttl > 1)
+    {
+      guint8 max = 1;
+      for (i = 0; i < udpsock->ttls->len; i++)
+      {
+        if (udpsock->ttls->data[i] > max)
+          max = udpsock->ttls->data[i];
+      }
+
+      if (max != udpsock->current_ttl)
+      {
+
+        if (setsockopt (udpsock->fd, IPPROTO_IP, IP_MULTICAST_TTL,
+                (const void *)&max, sizeof (max)) < 0)
+        {
+          GST_WARNING ("Error setting the multicast TTL to %u: %s", max,
+              g_strerror (errno));
+          g_mutex_unlock (trans->priv->mutex);
+          return;
+        }
+        udpsock->current_ttl = max;
+      }
+    }
+    g_mutex_unlock (trans->priv->mutex);
     return;
   }
 
   trans->priv->udpsocks[udpsock->component_id] =
     g_list_remove (trans->priv->udpsocks[udpsock->component_id], udpsock);
 
+  g_mutex_unlock (trans->priv->mutex);
+
   if (udpsock->udpsrc)
   {
     GstStateChangeReturn ret;
@@ -862,6 +955,13 @@ fs_multicast_transmitter_put_udpsock (FsMulticastTransmitter *trans,
     gst_object_unref (udpsock->udpsrc_requested_pad);
   }
 
+  if (udpsock->udpsink_requested_pad)
+  {
+    gst_element_release_request_pad (udpsock->tee,
+      udpsock->udpsink_requested_pad);
+    gst_object_unref (udpsock->udpsink_requested_pad);
+  }
+
   if (udpsock->udpsink)
   {
     GstStateChangeReturn ret;
@@ -874,44 +974,43 @@ fs_multicast_transmitter_put_udpsock (FsMulticastTransmitter *trans,
       GST_ERROR ("Could not remove udpsink element from transmitter source");
   }
 
-  if (udpsock->udpsink_requested_pad)
-  {
-    gst_element_release_request_pad (udpsock->tee,
-      udpsock->udpsink_requested_pad);
-    gst_object_unref (udpsock->udpsink_requested_pad);
-  }
-
   if (udpsock->fd >= 0)
     close (udpsock->fd);
 
+  g_byte_array_free (udpsock->ttls, TRUE);
+  g_free (udpsock->multicast_ip);
   g_free (udpsock->local_ip);
-  g_free (udpsock);
+  g_slice_free (UdpSock, udpsock);
 }
 
 void
 fs_multicast_transmitter_udpsock_inc_sending (UdpSock *udpsock)
 {
-  if (udpsock->sendcount == 0)
+  if (g_atomic_int_exchange_and_add (&udpsock->sendcount, 1) == 0)
     g_signal_emit_by_name (udpsock->udpsink, "add", udpsock->multicast_ip,
         udpsock->port);
-
-  udpsock->sendcount++;
 }
 
 void
 fs_multicast_transmitter_udpsock_dec_sending (UdpSock *udpsock)
 {
-  udpsock->sendcount--;
-
-  if (udpsock->sendcount == 0)
+  if (g_atomic_int_dec_and_test (&udpsock->sendcount))
     g_signal_emit_by_name (udpsock->udpsink, "remove", udpsock->multicast_ip,
         udpsock->port);
 }
 
 static GType
 fs_multicast_transmitter_get_stream_transmitter_type (
-    FsTransmitter *transmitter,
-    GError **error)
+    FsTransmitter *transmitter)
 {
   return FS_TYPE_MULTICAST_STREAM_TRANSMITTER;
 }
+
+void
+fs_multicast_transmitter_udpsock_ref (FsMulticastTransmitter *trans,
+    UdpSock *udpsock, guint8 ttl)
+{
+  g_mutex_lock (trans->priv->mutex);
+  g_byte_array_append (udpsock->ttls, &ttl, 1);
+  g_mutex_unlock (trans->priv->mutex);
+}
