@@ -42,11 +42,20 @@
 
 #include <string.h>
 #include <sys/types.h>
-#include <sys/socket.h>
-#include <netinet/in.h>
-#include <arpa/inet.h>
-#include <netdb.h>
-#include <unistd.h>
+
+#ifdef HAVE_UNISTD_H
+# include <unistd.h>
+#endif
+
+#ifdef G_OS_WIN32
+# include <winsock2.h>
+# define close closesocket
+#else /*G_OS_WIN32*/
+# include <netdb.h>
+# include <sys/socket.h>
+# include <netinet/in.h>
+# include <arpa/inet.h>
+#endif /*G_OS_WIN32*/
 
 GST_DEBUG_CATEGORY (fs_rawudp_transmitter_debug);
 #define GST_CAT_DEFAULT fs_rawudp_transmitter_debug
@@ -78,6 +87,8 @@ struct _FsRawUdpTransmitterPrivate
   GstElement **udpsrc_funnels;
   GstElement **udpsink_tees;
 
+  GMutex *mutex;
+  /* Protected by the mutex */
   GList **udpports;
 
   gboolean disposed;
@@ -109,8 +120,7 @@ static FsStreamTransmitter *fs_rawudp_transmitter_new_stream_transmitter (
     GParameter *parameters,
     GError **error);
 static GType fs_rawudp_transmitter_get_stream_transmitter_type (
-    FsTransmitter *transmitter,
-    GError **error);
+    FsTransmitter *transmitter);
 
 
 static GObjectClass *parent_class = NULL;
@@ -145,10 +155,9 @@ fs_rawudp_transmitter_register_type (FsPlugin *module)
     (GInstanceInitFunc) fs_rawudp_transmitter_init
   };
 
-  if (fs_rawudp_transmitter_debug == NULL)
-    GST_DEBUG_CATEGORY_INIT (fs_rawudp_transmitter_debug,
-        "fsrawudptransmitter", 0,
-        "Farsight raw UDP transmitter");
+  GST_DEBUG_CATEGORY_INIT (fs_rawudp_transmitter_debug,
+      "fsrawudptransmitter", 0,
+      "Farsight raw UDP transmitter");
 
   fs_rawudp_stream_transmitter_register_type (module);
 
@@ -158,18 +167,8 @@ fs_rawudp_transmitter_register_type (FsPlugin *module)
   return type;
 }
 
-static void
-fs_rawudp_transmitter_unload (FsPlugin *plugin)
-{
-  if (fs_rawudp_transmitter_debug)
-  {
-    gst_debug_category_free (fs_rawudp_transmitter_debug);
-    fs_rawudp_transmitter_debug = NULL;
-  }
-}
 
-FS_INIT_PLUGIN (fs_rawudp_transmitter_register_type,
-    fs_rawudp_transmitter_unload)
+FS_INIT_PLUGIN (fs_rawudp_transmitter_register_type)
 
 static void
 fs_rawudp_transmitter_class_init (FsRawUdpTransmitterClass *klass)
@@ -209,6 +208,7 @@ fs_rawudp_transmitter_init (FsRawUdpTransmitter *self)
   self->priv->disposed = FALSE;
 
   self->components = 2;
+  self->priv->mutex = g_mutex_new ();
 }
 
 static void
@@ -364,6 +364,8 @@ fs_rawudp_transmitter_constructed (GObject *object)
       return;
     }
   }
+
+  GST_CALL_PARENT (G_OBJECT_CLASS, constructed, (object));
 }
 
 static void
@@ -416,6 +418,8 @@ fs_rawudp_transmitter_finalize (GObject *object)
     self->priv->udpports = NULL;
   }
 
+  g_mutex_free (self->priv->mutex);
+
   parent_class->finalize (object);
 }
 
@@ -435,6 +439,9 @@ fs_rawudp_transmitter_get_property (GObject *object,
     case PROP_GST_SRC:
       g_value_set_object (value, self->priv->gst_src);
       break;
+    case PROP_COMPONENTS:
+      g_value_set_uint (value, self->components);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -494,6 +501,7 @@ fs_rawudp_transmitter_new_stream_transmitter (FsTransmitter *transmitter,
  */
 
 struct _UdpPort {
+  /* Protected by the transmitter mutex */
   gint refcount;
 
   GstElement *udpsrc;
@@ -514,6 +522,16 @@ struct _UdpPort {
   GstElement *tee;
 
   guint component_id;
+
+  /* Everything below is protected by the mutex */
+  GMutex *mutex;
+  GArray *known_addresses;
+};
+
+struct KnownAddress {
+  FsRawUdpAddressUniqueCallbackFunc callback;
+  gpointer user_data;
+  GstNetAddress addr;
 };
 
 static gint
@@ -541,11 +559,11 @@ _bind_port (
     retval = getaddrinfo (ip, NULL, &hints, &result);
     if (retval != 0)
     {
-      g_set_error (error, FS_ERROR, FS_ERROR_NETWORK,
+      g_set_error (error, FS_ERROR, FS_ERROR_INVALID_ARGUMENTS,
           "Invalid IP address %s passed: %s", ip, gai_strerror (retval));
       return -1;
     }
-    memcpy (&address, result->ai_addr, sizeof(struct sockaddr_in));
+    memcpy (&address, result->ai_addr, sizeof (struct sockaddr_in));
     freeaddrinfo (result);
   }
 
@@ -608,6 +626,11 @@ _create_sinksource (
       "sockfd", fd,
       NULL);
 
+  if (g_object_class_find_property (G_OBJECT_GET_CLASS (elem),
+          "auto-multicast"))
+    g_object_set (elem, "auto-multicast", FALSE, NULL);
+
+
   if (!gst_bin_add (bin, elem))
   {
     g_set_error (error, FS_ERROR, FS_ERROR_CONSTRUCTION,
@@ -677,24 +700,15 @@ _create_sinksource (
 }
 
 
-UdpPort *
-fs_rawudp_transmitter_get_udpport (FsRawUdpTransmitter *trans,
+static UdpPort *
+fs_rawudp_transmitter_get_udpport_locked (FsRawUdpTransmitter *trans,
     guint component_id,
     const gchar *requested_ip,
-    guint requested_port,
-    GError **error)
+    guint requested_port)
 {
   UdpPort *udpport;
   GList *udpport_e;
 
-  /* First lets check if we already have one */
-  if (component_id > trans->components)
-  {
-    g_set_error (error, FS_ERROR, FS_ERROR_INVALID_ARGUMENTS,
-        "Invalid component %d > %d", component_id, trans->components);
-    return NULL;
-  }
-
   for (udpport_e = g_list_first (trans->priv->udpports[component_id]);
        udpport_e;
        udpport_e = g_list_next (udpport_e))
@@ -711,16 +725,49 @@ fs_rawudp_transmitter_get_udpport (FsRawUdpTransmitter *trans,
     }
   }
 
+  return NULL;
+}
+
+
+UdpPort *
+fs_rawudp_transmitter_get_udpport (FsRawUdpTransmitter *trans,
+    guint component_id,
+    const gchar *requested_ip,
+    guint requested_port,
+    GError **error)
+{
+  UdpPort *udpport;
+  UdpPort *tmpudpport;
+
+  /* First lets check if we already have one */
+  if (component_id > trans->components)
+  {
+    g_set_error (error, FS_ERROR, FS_ERROR_INVALID_ARGUMENTS,
+        "Invalid component %d > %d", component_id, trans->components);
+    return NULL;
+  }
+
+  g_mutex_lock (trans->priv->mutex);
+  udpport = fs_rawudp_transmitter_get_udpport_locked (trans, component_id,
+      requested_ip, requested_port);
+  g_mutex_unlock (trans->priv->mutex);
+
+  if (udpport)
+    return udpport;
+
   GST_DEBUG ("Make new UdpPort for component %u requesting %s:%u", component_id,
       requested_ip ? requested_ip : "ANY", requested_port);
 
-  udpport = g_new0 (UdpPort, 1);
+  udpport = g_slice_new0 (UdpPort);
 
   udpport->refcount = 1;
   udpport->requested_ip = g_strdup (requested_ip);
   udpport->requested_port = requested_port;
   udpport->fd = -1;
   udpport->component_id = component_id;
+  udpport->mutex = g_mutex_new ();
+  udpport->known_addresses = g_array_new (TRUE, FALSE,
+      sizeof (struct KnownAddress));
 
   /* Now lets bind both ports */
 
@@ -751,8 +798,22 @@ fs_rawudp_transmitter_get_udpport (FsRawUdpTransmitter *trans,
       "sync", FALSE,
       NULL);
 
+  g_mutex_lock (trans->priv->mutex);
+
+  /* Check if someone else added the same port at the same time */
+  tmpudpport = fs_rawudp_transmitter_get_udpport_locked (trans, component_id,
+      requested_ip, requested_port);
+
+  if (tmpudpport)
+  {
+    g_mutex_unlock (trans->priv->mutex);
+    fs_rawudp_transmitter_put_udpport (trans, udpport);
+    return tmpudpport;
+  }
+
   trans->priv->udpports[component_id] =
     g_list_prepend (trans->priv->udpports[component_id], udpport);
+  g_mutex_unlock (trans->priv->mutex);
 
   return udpport;
 
@@ -768,15 +829,20 @@ fs_rawudp_transmitter_put_udpport (FsRawUdpTransmitter *trans,
 {
   GST_LOG ("Put port refcount %d->%d", udpport->refcount, udpport->refcount-1);
 
+  g_mutex_lock (trans->priv->mutex);
+
   if (udpport->refcount > 1)
   {
     udpport->refcount--;
+    g_mutex_unlock (trans->priv->mutex);
     return;
   }
 
   trans->priv->udpports[udpport->component_id] =
     g_list_remove (trans->priv->udpports[udpport->component_id], udpport);
 
+  g_mutex_unlock (trans->priv->mutex);
+
   if (udpport->udpsrc)
   {
     GstStateChangeReturn ret;
@@ -796,6 +862,13 @@ fs_rawudp_transmitter_put_udpport (FsRawUdpTransmitter *trans,
     gst_object_unref (udpport->udpsrc_requested_pad);
   }
 
+  if (udpport->udpsink_requested_pad)
+  {
+    gst_element_release_request_pad (udpport->tee,
+        udpport->udpsink_requested_pad);
+    gst_object_unref (udpport->udpsink_requested_pad);
+  }
+
   if (udpport->udpsink)
   {
     GstStateChangeReturn ret;
@@ -808,18 +881,16 @@ fs_rawudp_transmitter_put_udpport (FsRawUdpTransmitter *trans,
       GST_ERROR ("Could not remove udpsink element from transmitter source");
   }
 
-  if (udpport->udpsink_requested_pad)
-  {
-    gst_element_release_request_pad (udpport->tee,
-        udpport->udpsink_requested_pad);
-    gst_object_unref (udpport->udpsink_requested_pad);
-  }
-
   if (udpport->fd >= 0)
     close (udpport->fd);
 
+  if (udpport->mutex)
+    g_mutex_free (udpport->mutex);
+  if (udpport->known_addresses)
+    g_array_free (udpport->known_addresses, TRUE);
+
   g_free (udpport->requested_ip);
-  g_free (udpport);
+  g_slice_free (UdpPort, udpport);
 }
 
 void
@@ -912,8 +983,131 @@ fs_rawudp_transmitter_udpport_get_port (UdpPort *udpport)
 
 
 static GType
-fs_rawudp_transmitter_get_stream_transmitter_type (FsTransmitter *transmitter,
-    GError **error)
+fs_rawudp_transmitter_get_stream_transmitter_type (FsTransmitter *transmitter)
 {
   return FS_TYPE_RAWUDP_STREAM_TRANSMITTER;
 }
+
+/**
+ * fs_rawudp_transmitter_udpport_add_known_address:
+ * @udpport: a #UdpPort
+ * @address: the new #GstNetAddress that we know
+ * @callback: a Callback that will be called if the uniqueness of an address
+ *   changes
+ * @user_data: data passed back to the callback
+ *
+ * This function stores the passed address and tells the caller if it was
+ * unique or not. The callback is called when the uniqueness changes.
+ *
+ * Returns: %TRUE if the new address is unique, %FALSE otherwise
+ */
+
+gboolean
+fs_rawudp_transmitter_udpport_add_known_address (UdpPort *udpport,
+    GstNetAddress *address,
+    FsRawUdpAddressUniqueCallbackFunc callback,
+    gpointer user_data)
+{
+  gint i;
+  gboolean unique = FALSE;
+  struct KnownAddress newka = {0};
+  guint counter = 0;
+  struct KnownAddress *prev_ka = NULL;
+
+  g_mutex_lock (udpport->mutex);
+
+  for (i = 0;
+       g_array_index (udpport->known_addresses, struct KnownAddress, i).callback;
+       i++)
+  {
+    struct KnownAddress *ka = &g_array_index (udpport->known_addresses, struct KnownAddress, i);
+    if (gst_netaddress_equal (address, &ka->addr))
+    {
+      g_assert (!(ka->callback == callback && ka->user_data == user_data));
+
+      prev_ka = ka;
+      counter++;
+    }
+  }
+
+  if (counter == 0)
+  {
+    unique = TRUE;
+  }
+  else if (counter == 1)
+  {
+    if (prev_ka->callback)
+      prev_ka->callback (FALSE, &prev_ka->addr, prev_ka->user_data);
+  }
+
+  memcpy (&newka.addr, address, sizeof (GstNetAddress));
+  newka.callback = callback;
+  newka.user_data = user_data;
+
+  g_array_append_val (udpport->known_addresses, newka);
+
+  g_mutex_unlock (udpport->mutex);
+
+  return unique;
+}
+
+/**
+ * fs_rawudp_transmitter_udpport_remove_known_address:
+ * @udpport: a #UdpPort
+ * @address: the address to remove
+ * @callback: the callback passed to the corresponding
+ *  fs_rawudp_transmitter_udpport_add_known_address() call
+ * @user_data: the user_data passed to the corresponding
+ *  fs_rawudp_transmitter_udpport_add_known_address() call
+ *
+ * Removes a known address from the list and calls the notifiers if another
+ * address becomes unique
+ */
+
+void
+fs_rawudp_transmitter_udpport_remove_known_address (UdpPort *udpport,
+    GstNetAddress *address,
+    FsRawUdpAddressUniqueCallbackFunc callback,
+    gpointer user_data)
+{
+  gint i;
+  gint remove_i = -1;
+  guint counter = 0;
+  struct KnownAddress *prev_ka = NULL;
+
+  g_mutex_lock (udpport->mutex);
+
+  for (i = 0;
+       g_array_index (udpport->known_addresses, struct KnownAddress, i).callback;
+       i++)
+  {
+    struct KnownAddress *ka = &g_array_index (udpport->known_addresses, struct KnownAddress, i);
+    if (gst_netaddress_equal (address, &ka->addr))
+    {
+      if (ka->callback == callback && ka->user_data == user_data)
+      {
+        remove_i = i;
+      }
+      else
+      {
+        counter++;
+        prev_ka = ka;
+      }
+    }
+  }
+
+  if (remove_i == -1)
+  {
+    GST_ERROR ("Tried to remove unknown known address");
+    goto out;
+  }
+
+  if (counter == 1)
+    prev_ka->callback (TRUE, &prev_ka->addr, prev_ka->user_data);
+
+  g_array_remove_index_fast (udpport->known_addresses, remove_i);
+
+ out:
+
+  g_mutex_unlock (udpport->mutex);
+}
